with Flts;
use Flts;

pragma Elaborate_All (Flts);

generic

  MaxDeg: in Positive;  --- max degree

  type Scalar is private;
  type Scalars is array (Integer range <>) of Scalar;

  with function "="(S1,S2: Scalar) return Boolean is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with function Approx(S: Scalar) return Flt is <>;
  with function IsZero(S: Scalar) return Boolean is <>;

package ACM_RPoly is --- ACM Algorithm 493 by Jenkins and Traub

  subtype Coeffs is Scalars(1 .. MaxDeg+1); --- order of decreasing powers

  procedure RPoly(OP: in Coeffs; DEGREE: in out Integer; ZEROR,ZEROI: in out Scalars; FAIL: out Boolean);

private

  function ULP1 return Scalar;
  function Floor(S: in Scalar) return Integer;
  procedure FXSHFR(L2: in Integer; NZ: out Integer);
  procedure QUADIT(UU, VV: in Scalar; NZ: out Integer);
  procedure REALIT(SSS: in out Scalar; NZ,IFLAG: out Integer);
  procedure CALCSC(TYP: out Integer);
  procedure NEXTK(TYP: in Integer);
  procedure NEWEST(TYP: in Integer; UU,VV: in out Scalar);
  procedure QUADSD(NN0: in Integer; U0,V0: in Scalar; P0: in Coeffs; Q0: in out Coeffs; A0,B0: in out Scalar);
  procedure QUAD(A0,B1,C0: in Scalar; SR0,SI0,LR,LI: in out Scalar);

  Ten:    constant Flt := Flt(10);
  Twenty: constant Flt := Flt(20);
  ScalZero: constant Scalar := Scal(Zero);
  ScalHalf: constant Scalar := Scal(Half);
  ScalOne:  constant Scalar := Scal(One);

  BASE: constant Scalar := Scal(Two);
  LogBASE: constant Scalar := Log(BASE);
  ETA: constant Scalar := ULP1;
  SMALNO: constant Scalar := Scal(1.0E-999);  --- could be decreased
  INFIN: constant Scalar := Scal(1.0E+999);   --- could be increased

  N,NN: Integer;
  SR,SI,U,V,A,B,C,D,A1,A2,A3,A6,A7,E,F,G,H,SZR,SZI,LZR,LZI: Scalar;
  ARE: Scalar renames ETA;
  MRE: Scalar renames ETA;
  P,QP,K,QK,SVK: Coeffs;

end ACM_RPoly;
