with Ints, Flts, Globals, Reps, Params, Params.Names, Strings;
use Ints, Flts, Globals, Reps, Params, Strings;
with Matrices, Vectors, ScalVectors, CosSins1, CosSins1.Chain, CosSins1.Chain.Fix;

pragma Elaborate_All (Ints,Flts,Globals,Reps,Params.Names);
pragma Elaborate_All (Matrices,Vectors,ScalVectors,CosSins1,CosSins1.Chain,CosSins1.Chain.Fix);

procedure Approx_Fixpt(Sol: in Positive) is

  package PN is new Params.Names(Sol => Sol, Trunc => True);
  use PN;

  subtype Scalar is Rep;

  NumDigits: constant Positive := Reps.Set_Precision(128);

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SCS is new CosSins1 (FDeg => SizeCosSin, Scalar => Scalar, SV => SV);
  package SCSC is new SCS.Chain(JMax => SizeChain, Rho => RhoCh, Rad => RadCS, Phi1 => Phi1, Psi1 => Psi1, SM => SM);
  package Fix is new SCSC.Fix(Par => Par, Sigma => Sigma, Mat_PackSize => Mat_PackSize, TZ => TZ, TPi => TPi, Omega => Omega, SSV => SSV);

  use SM,SCS,SCSC,Fix;

  Eps: constant Radius := 1.0e-15;

  F: FChain;
  --  F1,F2: FChain;
  M: QC_Matrix;
  --  X: Flt := GetArg;
  --  Y: Flt := GetArg;
  --  Name: String := GetArg;

begin

  Read(Input_Dir & Fix_Name,F);
  FixRad(F);
  --  SetZero(Par,Sigma,F1);
  --  SetZero(Par,Sigma,F2);
  --  for J in -9..9 loop
  --     Copy(F.C(J-3),F1.C(J));
  --  end loop;
  --  for J in -9..9 loop
  --     Copy(F.C(J+3),F2.C(J));
  --  end loop;
  --  SetZero(Par,Sigma,F);
  --  Diff(F1,F2,F);
  --  Show1("",F);
  --  GMap(F,F1);
  --  Neg(F1);
  --  Add(F,F1);
  --  Show1("",MaxNorm(F1));
  --  Symmetrize_GMap := True;
  --  GMap(F,F1);
  --  Neg(F1);
  --  Add(F,F1);
  --  Show1("",MaxNorm(F1));
  --  for J in -4..0 loop
  --     Copy(F.C(J),F.C(J+3));
  --  end loop;
  --  SetZero(Par,Sigma,F);
  --  F.C(0).C(1) := Scal(X);
  --  F.C(0).C(1) := Scal(X);
  ApproxFix(F,M,Eps);
  --IterContr(F,M);
  Show1("norm F",MaxNorm(F));
  Write(Output_Dir & Mat_Name,M);
  Write(Output_Dir & Fix_Name,F);
  WritePlot(Output_Dir & Plot_Name,F,100);

end Approx_Fixpt;
