with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body CosSins1.Chain.Fix is

  use Lin,Lin_Contr;

  procedure SetZero(Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    Q.Sigma := Sigma;
    SetZero(Q.A);
    for J in Q.Sigma-JLast .. JLast loop SetZero(Rad,Par,C(J)); end loop;
  end SetZero;

  procedure Symmetrize(Q: in out FChain; Check: in Boolean := False) is
  begin
    Z_Symmetrize(TZ,Q);
    Pi_Symmetrize(TPi,Q,Check);
  end Symmetrize;

  procedure GMap(F: in FChain; GF: in out FChain) is
    --- GF := -InvHelm[ NablaStar(Phi'(Nabla F)) + Psi'(F)]
    use SV;
    CPhi1: constant SPoly1 := Convert(Phi1,1);
    CPsi1: constant SPoly1 := Convert(Psi1,2);
    Tmp: LF_Pointer;
  begin
    FPool.Allocate(Tmp);
    declare
      Q: FChain renames Tmp.all.Data;
    begin
      Nabla1(F,Q);
      Compose(CPhi1,Q,GF);
      NablaStar0(GF);        --- NablaStar(Phi'(Nabla F))
      if not IsZero(CPsi1) then
        Compose(CPsi1,F,Q);  --- Psi'(F)
        Add(Q,GF);
      end if;
    end;
    FPool.Recycle(Tmp);
    InvHelm(Scal_Omega,MM,GF);
    Neg(GF);
    if Symmetrize_GMap then Symmetrize(GF); end if;
  end GMap;

  procedure DGMap(Ph2,Ps2,H: in FChain; DGH: in out FChain) is
    --- DGH := -InvHelm[ NablaStar(Ph2*Nabla(H)) + Ps2*H]
    --- Ph2=Phi"(Nabla F)
    --- Ps2=Psi"(F)
    Tmp: LF_Pointer;
  begin
    FPool.Allocate(Tmp);
    declare
      Q: FChain renames Tmp.all.Data;
    begin
      Nabla1(H,Q);
      Prod(Ph2,Q,DGH);
      NablaStar0(DGH);    --- NablaStar(Ph2*Nabla(H))
      if not IsZero(Ps2) then
        Prod(Ps2,H,Q);    --- Ps2*H
        Add(Q,DGH);
      end if;
    end;
    FPool.Recycle(Tmp);
    InvHelm(Scal_Omega,MM,DGH);
    Neg(DGH);
    if Symmetrize_GMap then Symmetrize(DGH); end if;
  end DGMap;

  procedure DContrNorm(M: in SMatrix; F: in FChain; PackSize: in Positive; R: out Radius) is
    Q: constant CModes := Make(PackSize,Par,Sigma,True,False); --- Simple but with Err modes
    Ph2: constant FChain := Phi2(F,1);
    Ps2: constant FChain := Psi2(F,2);
    procedure DG(H: in FChain; DGH: in out FChain) is begin DGMap(Ph2,Ps2,H,DGH); end DG;
    procedure DContract is new DContr (DMap => DG);
    procedure DC(H: in FChain; DCH: in out FChain) is begin DContract(QC,M,H,DCH); end DC;
    procedure DCNorm is new Op_Norm (LinOp => DC);
  begin
    if PackSize <= Mat_PackSize then raise Not_Implemented; end if;
    if Verbosity>1 then
      Show1("DContrNorm: Q'Last = ",Q'Last);
    end if;
    DCNorm(Q,R);
  end DContrNorm;

  procedure AddSetB(U,V: in Scalar; Q: in out FChain) is
    RU: constant Radius := MaxNorm(U);
    RV: constant Radius := MaxNorm(V);
    C: CChain renames Q.C;
  begin
    if Not_STrunc then
      if RU /= Zero then
        for J in Sigma-JMax .. JMax loop
          AddBall(RU*InvRhoPow(abs(2*J-Sigma)),C(J));
        end loop;
      end if;
      if RV /= Zero then
        AddBall(RV*InvRhoPow(abs(2*JEMax-Sigma)),C(JEmax));
        AddBall(RV*InvRhoPow(abs(2*JEMax-Sigma)),C(Sigma-JEmax));
      end if;
    end if;
  end AddSetB;

  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out FChain; Multi: in Boolean) is
    KMax: constant Radius := Flt(7)/Flt(8);
    E,K: Radius;
    ANorm,R,Eps,Del: Scalar;
    F0: FChain;
  begin
    Show0("ContrFix: ",False);
    NumTrunc(FDeg,F,F0);
    GMap(F0,F);
    Sub(F0,F);
    E := MaxNorm(F);
    Assign(E,Eps);
    OpNorm(QC,M,ANorm);
    Add(SOne,ANorm);
    Show2("Eps,ANorm ",E,Sup(ANorm));
    ---
    Del := Eps/(SOne-Scal(KMax))+Scal(Tiny);
    R := ANorm*Del;
    Copy(F0,F);
    AddSetB(R,Two*Del,F);
    DContrNorm(M,F,PackSize,K);
    Show1("ContrFix: single K ",K);
    if K<=KMax and Multi then
      DContrMultiNorm(F,K);
      Show1("ContrFix:  multi K ",K);
    end if;
    if K>KMax then raise Sorry with "K too large"; end if;
    ---
    Del := Eps/(SOne-Scal(K))+Scal(Tiny);
    R := ANorm*Del;
    Show2("ContrFix: Del,R: ",Sup(Del),Sup(R));
    Copy(F0,F);
    AddSetB(R,Two*Del,F);
  end ContrFix;

  --------------------------------------------
  --------------------------------------------

  procedure DerMatrix(F: in FChain; M: in out SMatrix) is
    --- for numeric mode only
    Ph2: constant FChain := Phi2(F,1);
    Ps2: constant FChain := Psi2(F,2);
    procedure DG(H: in FChain; DGH: in out FChain) is begin DGMap(Ph2,Ps2,H,DGH); end DG;
    procedure Enclose is new Enclosure (LinOp => DG);
  begin
    if Not_STrunc then raise Sorry; end if;
    Enclose(QC,M);
  end DerMatrix;

  procedure ContrMatrix(F: in FChain; M: in out SMatrix) is
  begin
    if Verbosity>1 then Show0("running ContrMatrix"); end if;
    DerMatrix(F,M);
    Contr_Matrix(M);
  end ContrMatrix;

  procedure ApproxFix(F: in out FChain; M: in out SMatrix; Eps: in Radius := Zero) is
    --- find approximate fixed point
    procedure Contract0 is new Contr0 (Map => GMap);
    Iter: constant Integer := 1024;
    MaxCredit: constant Integer := 4;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := 1.0E+9;
    G: FChain;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running ApproxFix"); end if;
    Copy(F,G);
    Symmetrize_GMap := True;
    for I in 1 .. Iter loop
      ContrMatrix(G,M);
      Contract0(QC,M,G,Err);
      Show1("ApproxFix: step " & Strng(I) & ", Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        if (Err<Eps) or (Err=MinErr) then exit; end if;
        MinErr := Err;
        Credit := MaxCredit;
      else
        exit when Credit=0;
        Credit := Credit-1;
      end if;
    end loop;
    Show1("ApproxFix: Err ",Err);
    Symmetrize_GMap := False;
    ContrMatrix(F,M);
  end ApproxFix;

  procedure IterContr(F: in out FChain; M: in out SMatrix) is
    --- improve approximate fixed point
    procedure Contract0 is new Contr0 (Map => GMap);
    MaxCredit: constant Integer := 16;
    Credit: Integer := MaxCredit;
    Err,MinErr: Radius := One;
    G: FChain;
  begin
    if Not_STrunc then raise Sorry with "only for numeric mode"; end if;
    if Verbosity>1 then Show0("running IterContr"); end if;
    Copy(F,G);
    Symmetrize_GMap := True;
    ContrMatrix(G,M);
    while Credit>0 loop
      Contract0(QC,M,G,Err);
      Show1("IterContr: Err ",Err);
      if Err<MinErr then
        Copy(G,F);
        MinErr := Err;
        Credit := MaxCredit;
        ContrMatrix(G,M);
      else
        Credit := Credit-1;
      end if;
    end loop;
    Show1("IterContr: Err ",MinErr);
    Symmetrize_GMap := False;
    ContrMatrix(F,M);
  end IterContr;

  --------------------------------------------
  --------------------------------------------

  procedure DContrTailNorms(F: in FChain; L2L,R2R: out Radius) is
    Ph2: constant FChain := Phi2(F,1);
    Ps2: constant FChain := Psi2(F,2);
    Fac: constant Radius := InvRhoPow(abs(2*JEMax-Sigma));
    Tmp1,Tmp2: Flt;
    H,DGH: FChain;
    C: CChain renames H.C;
  begin
    --- left-tail, starts at Sigma-JMax-2
    SetZero(H);
    AddBall(Fac,C(Sigma-JEMax)); --- unit left-tail
    DGMap(Ph2,Ps2,H,DGH);
    MaxNorms(DGH,L2L,Tmp1,Tmp2);
    --- right-tail, starts at JMax+2
    SetZero(H);
    AddBall(Fac,C(JEMax));       --- unit right-tail
    DGMap(Ph2,Ps2,H,DGH);
    MaxNorms(DGH,Tmp1,Tmp2,R2R);
  end DContrTailNorms;

  procedure DContrMultiNorm(F: in FChain; K: in out Radius) is
    L2L,R2R,S1,S2,S3: Radius;
    C2L: Radius renames L2L;
    L2C: Radius renames L2L;
    C2C: Radius renames K;
    C2R: Radius renames R2R;
    R2C: Radius renames R2R;
  begin
    DContrTailNorms(F,L2L,R2R);
    S1 := L2L+C2L;              -- should be <= K/2
    S2 := L2C+C2C+R2C;          -- should be <= K
    S3 :=     C2R+R2R;          -- should be <= K/2
    K := RMax(Two*S1,S2,Two*S3);
    Show2("",L2l,R2r);
  end DContrMultiNorm;

end CosSins1.Chain.Fix;
