with ScalVectors, Linear, Linear.Contr;

pragma Elaborate_All (ScalVectors,Linear.Contr);

generic

  Par: in Parity;            --- 0: even in time   1: odd in time
  Sigma: in Parity;          --- 0: site-centered  1: bond-centered
  Mat_PackSize: in Positive;
  TZ:  in Symmetry;
  TPi: in Symmetry;
  Omega: in Rational;

  with procedure Put_Numeric(N: in Positive; S: in Scalar) is <>;
  with procedure Get_Numeric(N: in Positive; S: in out Scalar) is <>;

  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);

package CosSins1.Chain.Fix is

  Dummy: constant Boolean  := Set_CosSins1_Default(Rad,Par);

  --- for breathers
  procedure SetZero(Q: in out FChain);
  procedure Symmetrize(Q: in out FChain; Check: in Boolean := False);
  procedure GMap(F: in FChain; GF: in out FChain);
  procedure DGMap(Ph2,Ps2,H: in FChain; DGH: in out FChain);
  procedure DContrNorm(M: in SMatrix; F: in FChain; PackSize: in Positive; R: out Radius);
  procedure AddSetB(U,V: in Scalar; Q: in out FChain);
  procedure ContrFix(M: in SMatrix; PackSize: in Positive; F: in out FChain; Multi: in Boolean);

  package SV_Ops renames SSV.SV_Ops;
  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => FChain, Mode => CMode, Modes => CModes);
  package Lin_Contr is new Lin.Contr (SSV => SSV);

  --- numerical
  procedure DerMatrix(F: in FChain; M: in out SMatrix);
  procedure ContrMatrix(F: in FChain; M: in out SMatrix);
  procedure ApproxFix(F: in out FChain; M: in out SMatrix; Eps: in Radius := Zero);
  procedure IterContr(F: in out FChain; M: in out SMatrix);

  --- multi-breather stuff
  procedure DContrTailNorms(F: in FChain; L2L,R2R: out Radius);
  procedure DContrMultiNorm(F: in FChain; K: in out Radius);

  --- misc
  QC: constant CModes := Make(Mat_PackSize,Par,Sigma,True,True);   --- Simple and no Err modes
  subtype QC_Vector is SVector(QC'Range);
  subtype QC_Matrix is SMatrix(QC'Range,QC'Range);

  DContr_Parallel: Boolean renames Lin.Op_Norm_Parallel;
  DerMatrix_Parallel: Boolean renames Lin.Enclosure_Parallel;
  Symmetrize_GMap: Boolean := False; --- use only for numerical contraction

private

  Tiny: constant Radius := Flt'Scaling(One,Flt'Machine_Emin);
  MM: constant Scalar := Scal(-Psi1(1));
  Scal_Omega: constant Scalar := Scal(Omega);

end CosSins1.Chain.Fix;
