
package body CosSins1.Chain.FlokD is

  procedure MakeAB(Omega: in Rational; Q: in FChain; Alpha,Beta: in out FChain) is
    --- Alpha=Omega^{-2}Phi"(Nabla1 Q)
    --- Beta=Omega^{-2}Psi"(Q)
    OmegaM2: constant Scalar := Sqr(Scal(Inv(Omega)));
  begin
    Phi2(Q,Alpha);
    Mult(OmegaM2,Alpha);
    Psi2(Q,Beta);
    Mult(OmegaM2,Beta);
    if Q.Sigma/=Sigma or GetPar(Alpha,True)=1 or GetPar(Beta,True)=1 then raise Parity_Error; end if;
  end MakeAB;

  procedure CS2Mat(Alpha,Beta: in FChain; M1,M2: in out CSMat; Dm,Dp: out DVec) is
    use Flt_Vectors,SM;
    Pow2: constant Positive := 12;
    E: Radius;
    F,Sharp,Error: CosSin1;
  begin
    for K in 1 .. DimX loop         -- M2(*,K) := Beta(J) ~ Psi2(Q(J))   (sharpened)
      Component(K-JAst-1+Sigma,Beta,F); -- J=K-JAst-1+S
      SplitErr(F,Sharp,Error);
      for I in FFreq loop Copy(Sharp.C(I),M2(I,K)); end loop;
      Dm(K) := LInftyNorm(Error,Pow2);
    end loop;

    ZeroColumn(1,M1);               -- M1(*,1) := 0
    for K in 2 .. DimX loop         -- M1(*,K) := Alpha(J-S) ~ Phi2(Q(J)-Q(J-1))   (sharpened)
      Component(K-JAst-1,Alpha,F);
      SplitErr(F,Sharp,Error);
      for I in FFreq loop Copy(Sharp.C(I),M1(I,K)); end loop;
      E := Two*LInftyNorm(Error,Pow2);
      Dm(K-1) := Dm(K-1)+E;
      Dm(K) := Dm(K)+E;
    end loop;

    Dp := Dm;
    Component(-JAst,Alpha,F);
    Dp(1) := Dp(1)+Two*LInftyNorm(F,Pow2);
    Component(JAst-Sigma+1,Alpha,F);
    Dp(DimX) := Dp(DimX)+Two*LInftyNorm(F,Pow2);
  end CS2Mat;

  procedure High_D(Alpha,Beta: in FChain; DMin,DMax: out Flt) is
    --- bounds on d_j for J>JAst and J<(S-JAst)
    Pow2: constant Positive := 12;
    JLast1: constant Integer := JLast+1;
    Am0,Am1,Ap0,Ap1,Bm,Bp,Dm,Dp: Flt;
    Am,Ap: FltVec(-JLast1 .. JLast1) := (others => One);
    F: CosSin1;

    procedure RangePh2(J: in Integer; YMin,YMax: out Flt) is
    begin
      if Am(J)>Zero then
        Component(J,Alpha,F);    --- including Alpha.A
        YRange(F,YMin,YMax,Pow2);
        Am(J) := RMin(Zero,YMin);
        Ap(J) := RMax(Zero,YMax);
      end if;
      YMin := Am(J);
      YMax := Ap(J);
    end RangePh2;

    procedure RangePs2(J: in Integer; YMin,YMax: out Flt) is
    begin
      Component(J,Beta.Sigma,Beta.C,F); --- not including Beta.A
      YRange(F,YMin,YMax,Pow2);
    end RangePs2;

  begin
    DMin := Zero;
    DMax := Zero;
    for J in Sigma-1-JLast .. JLast+Sigma loop
      if J>JAst or J<(Sigma-JAst) then
        RangePh2(J-Sigma,Am0,Ap0);
        RangePh2(J-Sigma+1,Am1,Ap1);
        RangePs2(J,Bm,Bp);
        Dm := -(Two*(-Am0-Am1)-Bm); --- rounded down
        Dp := Two*(Ap0+Ap1)+Bp;     --- rounded up
        DMin := RMin(Dm,DMin);
        DMax := RMax(Dp,DMax);
      end if;
    end loop;
  end High_D;

  function High_Eta(BarBeta: Scalar; DMin,DMax: Flt) return F_Intervals is
    --- bounds on the high Floquet multipliers
    BMin: constant Radius := Inf(Sqrt(BarBeta+Scal(DMin)));
    BMax: constant Radius := Sup(Sqrt(BarBeta+Scal(DMax)));
    V: F_Intervals(1..2);
  begin
    V(1).Num_Lower := BMin;
    V(1).Num_Upper := BMax;
    V(1).Index := -1;
    V(1).Finite := False;
    V(2).Num_Lower := -(BMax-One);
    V(2).Num_Upper := One-BMin;
    V(2).Index := 1;
    V(2).Finite := False;
    return V;
  end High_Eta;

  function MaxNormHminusId(Alpha,Beta: FChain; Pow2: Positive := 12) return Radius is
    --- a tiny bit too large, since only H_0-Id is needed
    Lower,Upper: Flt;
    SMin,SMax,BetaMin,BetaMax: Scalar;
  begin
    YRange(Alpha,Lower,Upper,Pow2);
    Assign(Four*Lower,SMin);
    Assign(Four*Upper,SMax);
    YRange(Beta,Lower,Upper,Pow2);
    Assign(Lower,BetaMin);
    Assign(Upper,BetaMax);
    Add(BetaMin,SMin);
    Add(BetaMax,SMax);
    Sub(SOne,SMin);
    Sub(SOne,SMax);
    Lower := MaxNorm(SMin);
    Upper := MaxNorm(SMax);
    return RMax(Lower,Upper);
  end MaxNormHminusId;

  function MaxNormPhiDiff(Alpha,Beta: FChain; Dm,Dp: DVec; DMin,DMax: Flt) return Radius is
    --- bound on ||Phi-Phi_0||
    Pow2: constant Positive := 12;
    C: constant Radius := Half*MaxNormHminusId(Alpha,Beta,Pow2);
    Pi2: constant Scalar := Two*Pi;
    R: Radius := RMax(abs(DMin),abs(DMax));
  begin
    for K in 1 .. DimX loop
      R := RMax(Dm(K),R);
      R := RMax(Dp(K),R);
    end loop;
    if Verbosity>1 then
      Show1(" |d_j|  <",R);
      Show1("||I-H|| <",C);
    end if;
    R := MaxNorm(Pi2*Exp(C*Pi2))*R;
    return R;
  end MaxNormPhiDiff;

end CosSins1.Chain.FlokD;
