with Strings;
use Strings;

package body CosSins1.Chain.Pairs.FlokM is

  use SM,SV,SSV;

  procedure MakeD(Dm,Dp: in DVec; D: in out SVec) is
    K: Positive := 1;
    Sm,Sp: Scalar;
  begin
    for J in JUst .. JAst loop
      Assign(Half*Dp(K),Sp);
      Assign(Half*Dm(K),Sm);
      Sum(Sp,Sm,D(J));
      K := K+1;
    end loop;
  end MakeD;

  function MakeD(Dm,Dp: DVec) return SVec is
    D: SVec;
  begin
    MakeD(Dm,Dp,D);
    return D;
  end MakeD;

  procedure M2V(M: in CSMat; V: in out CSVec1) is
    K: Positive := 1;
  begin
    for J in JUst .. JAst loop
      declare
        VJ: CosSin1 renames V(J);
      begin
        VJ.R := Rad;
        VJ.P := 0;
        for I in FFreq loop Copy(M(I,K),VJ.C(I)); end loop;
        if Not_STrunc then VJ.E := (others => Zero); end if;
      end;
      K := K+1;
    end loop;
  end M2V;

  procedure Mat2CS(M1,M2: in CSMat; Dm,Dp: in DVec; AB: in out CSVec2) is
    --- convert matrices M1,M2 to chains AB(0),AB(1)
    A: CSVec1 renames AB(0);
    B: CSVec1 renames AB(1);
    K: Positive := 1;
    S,Tmp: Scalar;
  begin
    M2V(M1,A);                 --- A(J) := Alpha(J-S)   (sharpened)
    M2V(M2,B);                 --- B(J) := Beta(J)      (sharpened)
    for J in JUst .. JAst loop
      Assign(Half*Dp(K),S);    --- K = J+JAst+1-Sigma
      Assign(Half*Dm(K),Tmp);
      Sub(Tmp,S);
      Add(S,B(J).C(0));        --- add Half*(Dp-Dm) to B
      K := K+1;
    end loop;
  end Mat2CS;

  procedure NegH_Op(AB: in CSVec2; F: in CSVec1; G: in out CSVec1) is
    --- compare with MakeXh, see also CS2Mat
    Par: Parity renames F(1).P;
    A: CSVec1 renames AB(0); --- Alpha(J-Sigma) ~ Phi2(Q(J)-Q(J-1))
    B: CSVec1 renames AB(1); ---        Beta(J) ~ Psi2(Q(J))
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    declare
      U: CosSin1 renames Tmp.all.Data;
    begin
      for J in JUst .. JAst loop
        SetZero(Rad,Par,G(J));
        Copy(B(J),U);
        if J>JUst then
          Add(A(J),U);
          AddProd(A(J),F(J-1),G(J));   --- +Alpha(J-Sigma)*F(J-1)
        end if;
        if J<JAst then
          Add(A(J+1),U);
          AddProd(A(J+1),F(J+1),G(J)); --- +Alpha(J-Sigma+1)*F(J+1)
        end if;
        SubProd(U,F(J),G(J));          --- -[Beta(J)+Alpha(J-Sigma)+Alpha(J-Sigma+1)]*F(J)
      end loop;
    end;
    PPool.Recycle(Tmp);
  end NegH_Op;

  function InvTheta(I: Natural; J: Integer) return Positive is
    pragma Unreferenced(J);
  begin
    return IMax(I,1);
  end InvTheta;

  procedure Theta(V: in out CSVector) is
    Ti: Scalar;
  begin
    for J in V'Range loop
      declare
        P: CosSin1 renames V(J);
        C: FPoly1 renames P.C;
        E: Errs1 renames P.E;
      begin
        for I in 0 .. FDeg loop
          if not IsZero(C(I)) then
            Assign(InvTheta(I,J),Ti);
            Div(Ti,C(I));
          end if;
        end loop;
        if Not_Strunc then
          if P.P=1 and then (E(0)/=Zero) then -- LiftE0
            E(1) := E(1)+E(0)/P.R;
            E(0) := Zero;
          end if;
          for I in 0 .. EDeg loop
            if E(I)/=Zero then E(I) := E(I)/Flt(InvTheta(I,J)); end if;
          end loop;
        end if;
      end;
    end loop;
  end Theta;

  procedure Theta(V: in out CSVec2) is
  begin
    Theta(V(0));
    Theta(V(1));
  end Theta;

  procedure THT_Op(AB,F: in CSVec2; G: in out CSVec2) is
    --- Theta*(-H)*Theta
    use CSV;
    Tmp: LV_Pointer;
  begin
    VPool.Allocate(Tmp);
    declare
      FP: CSVec1 renames Tmp.all.Data;
    begin
      for P in Parity loop
        Copy(F(P),FP);
        Theta(FP);
        NegH_Op(AB,FP,G(P));
        Theta(G(P));
      end loop;
    end;
    VPool.Recycle(Tmp);
  end THT_Op;

  procedure THT_Matrix(AB: in CSVec2; THT: in out HMatrix) is
    --- Theta*(-H)*Theta
    use Lin;
    procedure THT_LinOp(F: in CSVec2; G: in out CSVec2) is begin THT_Op(AB,F,G); end THT_LinOp;
    procedure THT_Enclosure is new Enclosure (LinOp => THT_LinOp);
  begin
    THT_Enclosure(HMo,THT);
  end THT_Matrix;

  function HatMinusCheckMaxNorm2(THT: HMatrix) return Radius is
    Pow4: constant Positive := 4;
    Saved_MProd_Parallel: constant Boolean := MProd_Parallel;
    R: Radius;
    S: Scalar;
    HMC,Tmp: HMatrix;
  begin
    MProd_Parallel := MatProd_Parallel;
    Copy(THT,HMC);
    for M in CRange loop
      for N in CRange loop ModCenter(HMC(M,N)); end loop;
    end loop;
    MaxEigen(HMC,Tmp,S,Pow4);
    R := Sup(S);
    if Verbosity>1 then
      Show1("|hatH-checkH| <=",R);
    end if;
    MProd_Parallel := Saved_MProd_Parallel;
    return R;
  end HatMinusCheckMaxNorm2;

  procedure Low_L2(A: in out HMatrix) is
    --- also assumes/checks that A preserves parity
    SqrtTwo: constant Scalar := Sqrt(STwo);
    SqrtHalf: constant Scalar := Sqrt(SHalf);
    MFac,NFac: Integer;
  begin
    for M in HRange loop   --- cut and check
      for N in HRange loop
        if not IsZero(A(M,N)) then
          if HMo(M).Par /= HMo(N).Par then raise Parity_Error; end if;
          if M>CDim or N>CDim then SetZero(A(M,N)); end if;
        end if;
      end loop;
    end loop;
    for M in CRange loop   --- L2 basis
      if HMo(M).I=0 then MFac := 2; else MFac := 1; end if;
      for N in CRange loop
        if HMo(N).I=0 then NFac := 2; else NFac := 1; end if;
        if MFac>NFac then
          Mult(SqrtTwo,A(M,N));
        elsif MFac<NFac then
          Mult(SqrtHalf,A(M,N));
        end if;
      end loop;
    end loop;
  end Low_L2;

  function TMT_High(Eta: Scalar; D: SVec) return Flt is
    --- lower bound on Theta*[(I+Eta)^2-D(J)]*Theta for I>MaxSize
    I: constant Integer := MaxSize+1;
    Gap: Flt := One;
    B,Theta2: Scalar;
  begin
    if Sup(Eta)>=One then raise Not_Implemented; end if;
    for J in JUst .. JAst loop
      Theta2 := Inv(Scal(Sqr(InvTheta(I,J))));
      B := Theta2*(Sqr(Scal(I)+Eta)-D(J));
      Gap := RMin(Gap,Inf(B));
    end loop;
    return Gap;
  end TMT_High;

  procedure TMT_Low(Eta,S: in Scalar; D: in SVec; THT: in HMatrix; TMT: in out CMatrix) is
    --- notice that THT is Theta*(-H)*Theta
    Eta2: constant Scalar := Sqr(Eta);
    Nt: Natural;
    Theta2,B: Scalar;
  begin
    for N in CRange loop
      for M in CRange loop
        Center(THT(M,N),TMT(M,N));
      end loop;
      declare
        M: CMode renames CMo(N);
        Par: Parity renames M.Par;
        I: EFreq renames M.I;
        J: JRangeE renames M.J;
      begin
        Theta2 := Inv(Scal(Sqr(InvTheta(I,J))));
        B := Theta2*(Scal(I*I)+Eta2-S*D(J));
        Add(B,TMT(N,N));
        if I>0 then
          B := Theta2*Eta*Scal(2*I);
          Nt := Find(CMo,1-Par,I,J);
          if Nt=0 then raise Inconsistent_Data; end if;
          Add(B,TMT(Nt,N));
        end if;
      end;
    end loop;
  end TMT_Low;

  function HatNorm(D: SVec) return Radius is
    R: Radius := Zero;
  begin
    for J in D'Range loop
      R := RMax(MaxNorm(D(J))/Flt(Sqr(InvTheta(0,J))),R);
    end loop;
    return R;
  end HatNorm;

  procedure SpecGap(A,Tmp: in out CMatrix; Need: in Radius; OK: out Boolean) is
    --- modifies A and Tmp
    Saved_MProd_Parallel: constant Boolean := MProd_Parallel;
    Gap,Err: Radius;
    Eps: constant Radius := 1.0E-4;
  begin
    MProd_Parallel := MatProd_Parallel;
    Prod(A,A,Tmp);
    AddConst(Scal(-(Need*Need)),Tmp);
    OK := False;
    if IsPositive(Tmp)=True then
      Show0("gap test ok");
      OK := True;
    elsif Verbosity<2 then
      Show0("gap test failed");
    else
      NumSpecGap(A,Eps,Gap,Err);
      Show2("gap test failed. gap vs need: ",Gap,Need);
    end if;
    MProd_Parallel := Saved_MProd_Parallel;
  end SpecGap;

  procedure CheckEta(Eta: in Scalar; D: in SVec; THT: in HMatrix; Need: in Radius) is
    --- notice that THT is Theta*(-H)*Theta
    DS: constant Flt := Need*(-Two)/HatNorm(D);
    HiGap: constant Flt := TMT_High(Eta,D);
    OK: Boolean;
    TMT,Tmp: CMatrix;

    procedure CheckGap(S: in Flt) is
    begin
      Show1("S =",S,False);
      TMT_Low(Eta,Scal(S),D,THT,TMT);
      SpecGap(TMT,Tmp,Need,OK);
      if not OK then raise Sorry; end if;
    end CheckGap;

    S: Flt := DS;
  begin
    if Verbosity>2 then Show1("Checking eta =",Eta); end if;
    if HiGap<Need then
      Show2("high gap less than needed: ",HiGap,Need);
      raise Sorry;
    end if;
    CheckGap( One);
    CheckGap(Zero);
    while S>NegOne loop
      CheckGap( S);
      CheckGap(-S);
      S := S+DS;
    end loop;
    CheckGap(-One);
  end CheckEta;

  procedure CheckEta(Eta: in FltVec; D: in SVec; AB: in CSVec2) is
    Eps: constant Radius := 1.0E-999;
    Need: Radius;
    THT: HMatrix; --- Theta*(-H)*Theta
  begin
    Show0("CheckEta: checking " & Strng(Eta'Length) & " separating values");
    THT_Matrix(AB,THT);
    Need := HatMinusCheckMaxNorm2(THT)+Eps;
    Show1("need",Need);
    Low_L2(THT);
    if Verbosity>1 then Show1("THT symmetry:", SymmErr(THT)); end if;
    for N in Eta'Range loop
      if Verbosity>0 then Show1(RankStrng(N,2) & " eta:",Eta(N)); end if;
      if Eta(N)>Half then
        Show1("see ",One-Eta(N));
      else
        CheckEta(Scal(Eta(N)),D,THT,Need);
      end if;
    end loop;
    Show0("CheckEta: done");
  end CheckEta;

  procedure Show_TMT_Low_Gaps(S: in Scalar; D: in SVec; AB: in CSVec2) is
    Steps: constant Positive := 100;
    Eps: constant Radius := 1.0E-4;
    Eta,Gap,Err: Flt;
    TMT: CMatrix;
    THT: HMatrix;
  begin
    THT_Matrix(AB,THT);
    for N in 0 .. Steps loop
      Eta := Half*Flt(N)/Flt(Steps);
      TMT_Low(Scal(Eta),S,D,THT,TMT);
      NumSpecGap(TMT,Eps,Gap,Err);
      Show2("Eta Gap ",Eta,Gap);
    end loop;
  end Show_TMT_Low_Gaps;

end CosSins1.Chain.Pairs.FlokM;
