with ScalVectors, Recycling, Linear;

pragma Elaborate_All (ScalVectors,Recycling,Linear);

generic

  MaxSize: in Positive;

  with procedure Put_Numeric(N: in Positive; S: in Scalar) is <>;
  with procedure Get_Numeric(N: in Positive; S: in out Scalar) is <>;

  with package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);

package CosSins1.Chain.Pairs.FlokM is

  subtype DVec is FltVec(1 .. DimX);
  subtype SVec is SVector(JUst .. JAst);
  subtype CSMat is SMatrix(FFreq,1 .. DimX);

  procedure MakeD(Dm,Dp: in DVec; D: in out SVec);
  function MakeD(Dm,Dp: DVec) return SVec;
  procedure Mat2CS(M1,M2: in CSMat; Dm,Dp: in DVec; AB: in out CSVec2);

  --- \hat stuff
  HMo: constant CModes := Make(MaxSize,False,False);
  HDim: constant Positive  := HMo'Last;
  subtype HRange is Integer range 1 .. HDim;    -- for \hat matrices
  subtype HVector is SVector(HRange);
  subtype HMatrix is SMatrix(HRange,HRange);

  function InvTheta(I: Natural; J: Integer) return Positive;
  procedure Theta(V: in out CSVector);
  procedure Theta(V: in out CSVec2);
  procedure THT_Op(AB,F: in CSVec2; G: in out CSVec2);
  procedure THT_Matrix(AB: in CSVec2; THT: in out HMatrix);

  --- \check stuff
  CDim: constant Positive := HDim-NumErrModes;
  CMo: constant CModes := HMo(1 .. CDim);
  subtype CRange is Integer range 1 .. CDim;    -- for \check matrices
  subtype CVector is SVector(CRange);
  subtype CMatrix is SMatrix(CRange,CRange);

  function HatMinusCheckMaxNorm2(THT: HMatrix) return Radius;
  function TMT_High(Eta: Scalar; D: SVec) return Flt;
  procedure TMT_Low(Eta,S: in Scalar; D: in SVec; THT: in HMatrix; TMT: in out CMatrix);
  procedure CheckEta(Eta: in Scalar; D: in SVec; THT: in HMatrix; Need: in Radius);
  procedure CheckEta(Eta: in FltVec; D: in SVec; AB: in CSVec2);

  package Lin is new Linear (Scalar => Scalar, SM => SM, SV => SV, Fun => CSVec2, Mode => CMode, Modes => CModes);

  --- purely numerical
  procedure Show_TMT_Low_Gaps(S: in Scalar; D: in SVec; AB: in CSVec2);

  MatProd_Parallel: Boolean := True;
  Op_Norm_Parallel: Boolean renames Lin.Op_Norm_Parallel;
  Enlosure_Parallel: Boolean renames Lin.Enclosure_Parallel;

private

  procedure M2V(M: in CSMat; V: in out CSVec1);
  procedure NegH_Op(AB: in CSVec2; F: in CSVec1; G: in out CSVec1);
  procedure Low_L2(A: in out HMatrix);
  function HatNorm(D: SVec) return Radius;
  procedure SpecGap(A,Tmp: in out CMatrix; Need: in Radius; OK: out Boolean);

  package VAlloc is new Recycling (Item => CSVec1);
  subtype LV_Pointer is VAlloc.LD_Pointer;
  VAC: VAlloc.Controlled_Protected_Data;
  VPool: VAlloc.Protected_Data renames VAC.P;

end CosSins1.Chain.Pairs.FlokM;
