
package body CosSins1.Chain.Pairs is

  procedure SetZero(V: in out CSVec2) is
    V0: CSVec1 renames V(0);
    V1: CSVec1 renames V(1);
  begin
    for J in V0'Range loop
      SetZero(Rad,0,V0(J));
      SetZero(Rad,1,V1(J));
    end loop;
  end SetZero;

  procedure Norm(V: in CSVec2; N: in out Scalar) is
    V0: CSVec1 renames V(0);
    V1: CSVec1 renames V(1);
    NJ,Tmp: Scalar;
  begin
    SetZero(N);
    for J in V0'Range loop
      Norm(V0(J),NJ);
      Norm(V1(J),Tmp);
      Add(Tmp,NJ);
      AddProd(RhoPower(abs(2*J-Sigma)),NJ,N);
    end loop;
  end Norm;

  function Norm(V: CSVec2) return Scalar is
    N: Scalar;
  begin
    Norm(V,N);
    return N;
  end Norm;

  function MaxNorm(V: CSVec2) return Radius is
    V0: CSVec1 renames V(0);
    V1: CSVec1 renames V(1);
    R: Radius := Zero;
  begin
    for J in V0'Range loop
      R := R+RhoPow(abs(2*J-Sigma))*(MaxNorm(V0(J))+MaxNorm(V1(J)));
    end loop;
    return R;
  end MaxNorm;

  procedure Show1(N: in String; V: in CSVec2; Hide0: in Boolean := True) is
    use CSV;
  begin
    Show1(N & "0 ",V(0),Hide0);
    Show1(N & "1 ",V(1),Hide0);
  end Show1;

  procedure ParZero(Par: in Parity; V: in out CSVec2) is
    VP: CSVec1 renames V(Par);
  begin
    for J in VP'Range loop SetZero(Rad,Par,VP(J)); end loop;
  end ParZero;

  procedure Assign(M: in CMode; V: in out CSVec2) is
  begin
    Assign(M,V(M.Par));
    ParZero(1-M.Par,V);
  end Assign;

  procedure AddProd(M: in CMode; S: in Scalar; V: in out CSVec2; Check: in Boolean) is
    pragma Unreferenced(Check);
  begin
    AddProd(M,S,V(M.Par));
  end AddProd;

  procedure Extract(M: in CMode; V: in out CSVec2; S: in out Scalar; Check: in Boolean) is
    pragma Unreferenced(Check);
  begin
    Extract(M,V(M.Par),S);
    if (M.Sigma/=Sigma) then raise Parity_Error; end if;
    --- maybe check for partition as well
  end Extract;

  procedure ErrsLast(M: in out CModes) is
    --- use only if M includes error modes
    KC,KE: Positive := 1;
    MC: CModes(1 .. M'Last-NumErrModes);
    ME: CModes(1 .. NumErrModes);
  begin
    for K in M'Range loop
      if M(K).Err then
        ME(KE) := M(K);
        KE := KE+1;
      else
        MC(KC) := M(K);
        KC := KC+1;
      end if;
    end loop;
    M := MC & ME;
  end ErrsLast;

  function Make(Size: FFreq; Simple,Trunc: Boolean) return CModes is
    --- error modes are at the end
    Simp: constant Boolean := Simple or Trunc or STrunc;
    Include_Errs: constant Boolean := not (Trunc or STrunc);
    function MaxDeg(N: Natural) return Integer is
      pragma Unreferenced(N);
    begin
      return Size;
    end MaxDeg;
    function GMake is new Gen_Make (MaxDeg => MaxDeg);
  begin
    if JAst >= JMax then raise Not_Implemented; end if;
    declare
      M0: constant CModes := GMake(JAst,0,Sigma,Simp,Include_Errs);
      M1: constant CModes := GMake(JAst,1,Sigma,Simp,Include_Errs);
      M: CModes(1 .. M0'Last+M1'Last) := M0 & M1;
    begin
      if Include_Errs then ErrsLast(M); end if;
      return M;
    end;
  end Make;

end CosSins1.Chain.Pairs;
