with Strings, Protected_Counters;
use Strings, Protected_Counters;

pragma Elaborate_All (Strings, Protected_Counters);

package body CosSins1.Chain is

  use SV,CSV;

  procedure SetZero(Par,Sigma: in Parity; Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    Q.Sigma := Sigma;
    SetZero(Q.A);
    for J in Sigma-JLast .. JLast loop SetZero(Rad,Par,C(J)); end loop;
  end SetZero;

  function CheckRad(Q: FChain) return Boolean is
    --- check if all CosSin1 radii are Rad
    C: CChain renames Q.C;
  begin
    for J in Q.Sigma-JLast .. JLast loop
      if C(J).R /= Rad then return False; end if;
    end loop;
    return True;
  end CheckRad;

  procedure FixRad(Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    for J in Q.Sigma-JLast .. JLast loop
      if C(J).R >= Rad or else IsPoly(C(J)) then
        C(J).R := Rad;
      else
        raise Domain_Error with "cannot increase domain";
      end if;
    end loop;
  end FixRad;

  function GetPar(Q: FChain; Check: Boolean := False) return Parity is
    C: CChain renames Q.C;
    P: Parity renames C(0).P;
  begin
    if Check then
      if not (P=0 or IsZero(Q.A)) then raise Domain_Error; end if;
      for J in Q.Sigma-JLast .. JLast loop
        if C(J).P /= P then raise Domain_Error; end if;
      end loop;
    end if;
    return P;
  end GetPar;

  procedure Copy(Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    Q2.Sigma := Q1.Sigma;
    Copy(Q1.A,Q2.A);
    for J in Q1.Sigma-JLast .. JLast loop Copy(C1(J),C2(J)); end loop;
  end Copy;

  function IsZero(Q: FChain) return Boolean is
    C: CChain renames Q.C;
  begin
    if not IsZero(Q.A) then return False; end if;
    for J in Q.Sigma-JLast .. JLast loop
      if not IsZero(C(J)) then return False; end if;
    end loop;
    return True;
  end IsZero;

  procedure Neg(Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    Neg(Q.A);
    for J in Q.Sigma-JLast .. JLast loop Neg(C(J)); end loop;
  end Neg;

  procedure Neg(Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    Q2.Sigma := Q1.Sigma;
    Neg(Q1.A,Q2.A);
    for J in Q1.Sigma-JLast .. JLast loop Neg(C1(J),C2(J)); end loop;
  end Neg;

  procedure AddErrs(P1: in CosSin1; P2: in out CosSin1) is
    --- use only if Not_STrunc
    E1: Errs1 renames P1.E;
    E2: Errs1 renames P2.E;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    if P2.R>P1.R then P2.R := P1.R; end if;
    for I in EFreq loop E2(I) := E2(I)+E1(I); end loop;
  end AddErrs;

  procedure Add(Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    if Q1.Sigma /= Q2.Sigma then raise Parity_Error; end if;
    Add(Q1.A,Q2.A);
    for J in Q1.Sigma-JMax .. JMax loop Add(C1(J),C2(J)); end loop;
    if Not_STrunc then
      AddErrs(C1(JEMax),C2(JEMax));
      AddErrs(C1(Q1.Sigma-JEMax),C2(Q1.Sigma-JEMax));
    end if;
  end Add;

  procedure SumErrs(P1,P2: in CosSin1; P3: in out CosSin1) is
    --- use only if Not_Strunc
    E1: Errs1 renames P1.E;
    E2: Errs1 renames P2.E;
    E3: Errs1 renames P3.E;
    C3: FPoly1 renames P3.C;
  begin
    P3.R := RMin(P1.R,P2.R);
    if P1.P /= P2.P then raise Parity_Error; end if;
    P3.P := P1.P;
    for I in P1.P .. FDeg loop SetZero(C3(I)); end loop;
    for I in EFreq loop E3(I) := E1(I)+E2(I); end loop;
  end SumErrs;

  procedure Sum(Q1,Q2: in FChain; Q3: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
    C3: CChain renames Q3.C;
  begin
    if Q1.Sigma /= Q2.Sigma then raise Parity_Error; end if;
    Q3.Sigma := Q1.Sigma;
    Sum(Q1.A,Q2.A,Q3.A);
    for J in Q1.Sigma-JMax .. JMax loop Sum(C1(J),C2(J),C3(J)); end loop;
    if Not_STrunc then
      SumErrs(C1(JEMax),C2(JEMax),C3(JEMax));
      SumErrs(C1(Q1.Sigma-JEMax),C2(Q1.Sigma-JEMax),C3(Q1.Sigma-JEMax));
    end if;
  end Sum;

  procedure Sub(Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    if Q1.Sigma /= Q2.Sigma then raise Parity_Error; end if;
    Sub(Q1.A,Q2.A);
    for J in Q1.Sigma-JMax .. JMax loop Sub(C1(J),C2(J)); end loop;
    if Not_STrunc then
      AddErrs(C1(JEMax),C2(JEMax));
      AddErrs(C1(Q1.Sigma-JEMax),C2(Q1.Sigma-JEMax));
    end if;
  end Sub;

  procedure Diff(Q1,Q2: in FChain; Q3: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
    C3: CChain renames Q3.C;
  begin
    if Q1.Sigma /= Q2.Sigma then raise Parity_Error; end if;
    Q3.Sigma := Q1.Sigma;
    Diff(Q1.A,Q2.A,Q3.A);
    for J in Q1.Sigma-JMax .. JMax loop Diff(C1(J),C2(J),C3(J)); end loop;
    if Not_STrunc then
      SumErrs(C1(JEMax),C2(JEMax),C3(JEMax));
      SumErrs(C1(Q1.Sigma-JEMax),C2(Q1.Sigma-JEMax),C3(Q1.Sigma-JEMax));
    end if;
  end Diff;

  procedure Mult(R: in Flt; Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    Mult(R,Q.A);
    for J in Q.Sigma-JLast .. JLast loop Mult(R,C(J)); end loop;
  end Mult;

  procedure Prod(R: in Flt; Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    Q2.Sigma := Q1.Sigma;
    Prod(R,Q1.A,Q2.A);
    for J in Q1.Sigma-JLast .. JLast loop Prod(R,C1(J),C2(J)); end loop;
  end Prod;

  function "*"(R: Flt; Q: FChain) return FChain is
    Q1: FChain;
  begin
    Prod(R,Q,Q1);
    return Q1;
  end "*";

  procedure Mult(S: in Scalar; Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    Mult(S,Q.A);
    for J in Q.Sigma-JLast .. JLast loop Mult(S,C(J)); end loop;
  end Mult;

  procedure Prod(S: in Scalar; Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    Q2.Sigma := Q1.Sigma;
    Prod(S,Q1.A,Q2.A);
    for J in Q1.Sigma-JLast .. JLast loop Prod(S,C1(J),C2(J)); end loop;
  end Prod;

  function "*"(S: Scalar; Q: FChain) return FChain is
    Q1: FChain;
  begin
    Prod(S,Q,Q1);
    return Q1;
  end "*";

  procedure AddProd(S: in Scalar; Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    if not IsZero(S) then
      if Q1.Sigma /= Q2.Sigma then raise Parity_Error; end if;
      AddProd(S,Q1.A,Q2.A);
      for J in Q1.Sigma-JLast .. JLast loop AddProd(S,C1(J),C2(J)); end loop;
    end if;
  end AddProd;

  procedure AddProd(S: in Scalar; C1: in CChain; Q2: in out FChain) is
    C2: CChain renames Q2.C;
  begin
    if not IsZero(S) then
      for J in Q2.Sigma-JLast .. JLast loop AddProd(S,C1(J),C2(J)); end loop;
    end if;
  end AddProd;

  procedure Prod(Q1,Q2: in FChain; Q3: in out FChain) is
    --- pointwise multiplication
    JEMin: constant Integer := Q1.Sigma-JEMax;
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
    C3: CChain renames Q3.C;
    Task_Error: Boolean := False;
    PC: Protected_Counter(JEMin);

    procedure Prod_Loop is
      J: Integer;
    begin
      Proper_Rounding;
      loop
        PC.Next(J);
        exit when J>JEMax;
        if Not_STrunc or else (J>JEMin and then J<JEMax) then
          Prod(C1(J),C2(J),C3(J));
        end if;
      end loop;
    exception
      when others => Task_Error := True; raise;
    end Prod_Loop;

    NT: constant Natural := Reserve_Tasks(2*JMax,Prod_Parallel);
  begin
    if Q1.Sigma /= Q2.Sigma then raise Parity_Error; end if;
    Q3.Sigma := Q1.Sigma;
    Prod(Q1.A,Q2.A,Q3.A);
    if NT=0 then
      Prod_Loop;
    else
      declare
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is begin Prod_Loop; end Prod_Task_Type;
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
      begin
        Prod_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
    AddProd(Q1.A,C2,Q3);
    AddProd(Q2.A,C1,Q3);
  end Prod;

  procedure Norm(Q: in FChain; S: in out Scalar) is
    Sigma: Parity renames Q.Sigma;
    C: CChain renames Q.C;
    SJ: Scalar;
  begin
    if not IsZero(Q.A) then raise Undefined; end if;
    SetZero(S);
    for J in Sigma-JLast .. JLast loop
      Norm(C(J),SJ);
      AddProd(RhoPower(abs(2*J-Sigma)),SJ,S);
    end loop;
  end Norm;

  function MaxNorm(Q: FChain) return Radius is
    Sigma: Parity renames Q.Sigma;
    C: CChain renames Q.C;
    R: Radius := Zero;
  begin
    if not IsZero(Q.A) then raise Undefined; end if;
    for J in Sigma-JLast .. JLast loop
      R := R+RhoPow(abs(2*J-Sigma))*MaxNorm(C(J));
    end loop;
    return R;
  end MaxNorm;

  procedure MaxNorms(Q: in FChain; R1,R2,R3: out Radius) is
    --- MaxNorm of left-tail, center, right-tail
    Sigma: Parity renames Q.Sigma;
    QC: CChain renames Q.C;
  begin
    if not IsZero(Q.A) then raise Undefined; end if;
    R2 := Zero;
    for J in Sigma-JMax .. JMax loop
      R2 := R2+RhoPow(abs(2*J-Sigma))*MaxNorm(QC(J));
    end loop;
    if STrunc then
      R1 := Zero;
      R2 := Zero;
    else
      R1 := RhoPow(abs(2*JEMax-Sigma))*MaxNorm(QC(Sigma-JEMax));
      R3 := RhoPow(abs(2*JEMax-Sigma))*MaxNorm(QC(JEMax));
    end if;
  end MaxNorms;

  procedure YRange(Q: in FChain; YInf,YSup: out Flt; Pow2: in Positive := 12) is
    C: CChain renames Q.C;
    YMin,YMax: Flt;
  begin
    YRange(C(JLast),YInf,YSup,Pow2);
    for J in Q.Sigma-JLast .. JLast-1 loop
      YRange(C(J),YMin,YMax,Pow2);
      YInf := RMin(YMin,YInf);
      YSup := RMax(YMax,YSup);
    end loop;
    if not IsZero(Q.A) then
      YSup := YSup+Sup(Q.A);
      YInf := -(-YInf-Inf(Q.A));
    end if;
  end YRange;

  function Inf(Q: FChain; Pow2: Positive := 12) return Flt is
    YInf,YSup: Flt;
  begin
    YRange(Q,YInf,YSup,Pow2);
    return YInf;
  end Inf;

  function Sup(Q: FChain; Pow2: Positive := 12) return Flt is
    YInf,YSup: Flt;
  begin
    YRange(Q,YInf,YSup,Pow2);
    return YSup;
  end Sup;

  procedure NumTrunc(Siz: in Natural; Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    Q2.Sigma := Q1.Sigma;
    Center(Q1.A,Q2.A);
    for J in Q1.Sigma-JMax .. JMax loop NumTrunc(Siz,C1(J),C2(J)); end loop;
    if Not_STrunc then
      SetZero(C2(JEMax));
      SetZero(C2(Q1.Sigma-JEMax));
    end if;
  end NumTrunc;

  procedure AddBall(R: in Radius; Q: in out FChain) is
    Sigma: Parity renames Q.Sigma;
    C: CChain renames Q.C;
  begin
    if not (STrunc or else R=Zero) then
      for J in Sigma-JEMax .. JEMax loop
        AddBall(R*InvRhoPow(abs(2*J-Sigma)),C(J));
      end loop;
    end if;
  end AddBall;

  procedure AddBall(S: in Scalar; Q: in out FChain) is
  begin
    if Not_STrunc then AddBall(MaxNorm(S),Q); end if;
  end AddBall;

  procedure Component(J: in Integer; Sigma: in Parity; C: in CChain; P: in out CosSin1) is
    N: constant Natural := IMax(J,Sigma-J);
  begin
    if N <= JMax then                                  --- plain coeff site
      Copy(C(J),P);
    elsif STrunc then
      SetZero(C(1).R,C(1).P,P);
    elsif J >= JEMax then                              --- right tail
      Prod(NPower(InvRho2,J-JEMax),C(JEMax),P);
    else                                               --- left tail
      Prod(NPower(InvRho2,N-JEMax),C(Sigma-JEMax),P);
    end if;
  end Component;

  procedure Component(J: in Integer; Q: in FChain; P: in out CosSin1) is
  begin
    Component(J,Q.Sigma,Q.C,P);
    AddCoeff(0,Q.A,P);
  end Component;

  procedure Add_Site(J: in Integer; P: in CosSin1; Q: in out FChain) is
    N: constant Natural := IMax(J,Q.Sigma-J);
    C: CChain renames Q.C;
  begin
    if P.P /= C(0).P then raise Undefined; end if;
    if Not_STrunc and then (P.R<Rad) then raise Undefined; end if;
    if N <= JMax then                                  --- plain coeff site
      Add(P,C(J));
    elsif STrunc then
      null;
    elsif J >= JEMax then                              --- right tail
      AddProd(NPower(Rho2,J-JEMax),P,C(JEMax));
    else                                               --- left tail
      AddProd(NPower(Rho2,N-JEMax),P,C(Q.Sigma-JEMax));
    end if;
  end Add_Site;

  procedure Add_Tail(J: in Integer; P: in CosSin1; Q: in out FChain) is
    --- P is assumed to be a tail starting from J
    N: constant Natural := IMax(J,Q.Sigma-J);
    C: CChain renames Q.C;
    R: Flt := One;
  begin
    if STrunc then return; end if;
    if P.P /= C(0).P or (P.R<Rad) then raise Undefined; end if;
    if J>JEMax then
      AddProd(NPower(Rho2,J-JEMax),P,C(JEMax));
    elsif N>JEMax then
      AddProd(NPower(Rho2,N-JEMax),P,C(Q.Sigma-JEMax));
    elsif N=J then
      if J=0 then raise Undefined; end if;
      for K in J .. JEMax loop             --- right tail
        AddProd(R,P,C(K));
        R := R*InvRho2;
      end loop;
    else
      for K in reverse Q.Sigma-JEMax .. J loop
        AddProd(R,P,C(K));
        R := R*InvRho2;
      end loop;
    end if;
  end Add_Tail;

  procedure SetAverage0(Q: in out FChain) is
    C: CChain renames Q.C;
    Err: Flt := Zero;
    S: Scalar;
  begin
    if not IsZero(Q.A) then raise Not_Implemented; end if;
    for J in Q.Sigma-JMax .. JMax loop
      SplitOffAverage(C(J),S);
      Err := RMax(Err,Inf(S),-Sup(S));
    end loop;
    if Verbosity>1 and then Err>Zero then
      Show1("average correction >",Err);
    end if;
    if Not_STrunc then
      LiftE0(C(JEMax));
      LiftE0(C(Q.Sigma-JEMax));
    end if;
  end SetAverage0;

  procedure Project(J: in Integer; Q: in out FChain) is
    JFirst: constant Integer := Q.Sigma-JLast;
    J1: constant Integer := IMax(JFirst,IMin(JLast,J));
    C: CChain renames Q.C;
  begin
    for J2 in JFirst .. JLast loop
      if J2 /= J1 then FillZero(C(J2)); end if;
    end loop;
  end Project;

  procedure Reflect(Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    for J in 1 .. JLast loop Swap(C(Q.Sigma-J),C(J)); end loop;
  end Reflect;

  procedure Reflect(Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    Q2.Sigma := Q1.Sigma;
    Copy(Q1.A,Q2.A);
    for J in Q1.Sigma-JLast .. JLast loop Copy(C1(Q1.Sigma-J),C2(J)); end loop;
  end Reflect;

  procedure TranslatePi(Q: in out FChain) is
    C: CChain renames Q.C;
  begin
    for J in Q.Sigma-JLast .. JLast loop TranslatePi(C(J)); end loop;
  end TranslatePi;

  procedure TranslatePi(Q1: in FChain; Q2: in out FChain) is
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
  begin
    Q2.Sigma := Q1.Sigma;
    Copy(Q1.A,Q2.A);
    for J in Q1.Sigma-JLast .. JLast loop TranslatePi(C1(J),C2(J)); end loop;
  end TranslatePi;

  procedure Z_Symmetrize(T: in Symmetry; Q: in out FChain) is
    --- reflection with time-translation zero
    C: CChain renames Q.C;
  begin
    if T=Symm then
      for J in 1 .. JLast loop
        Add(C(Q.Sigma-J),C(J));
        Mult(Half,C(J));
        Copy(C(J),C(Q.Sigma-J));
      end loop;
    elsif T=ASymm then
      SetZero(Q.A);
      if Q.Sigma=0 then SetZero(C(0)); end if;
      for J in 1 .. JLast loop
        Sub(C(Q.Sigma-J),C(J));
        Mult(Half,C(J));
        Neg(C(J),C(Q.Sigma-J));
      end loop;
    end if;
  end Z_Symmetrize;

  procedure Pi_Symmetrize(T: in Symmetry; Q: in out FChain; Check: in Boolean := False) is
    C: CChain renames Q.C;
  begin
    if T=Symm then
      if Q.Sigma=0 then SymmetrizePi(0,C(0),Check); end if;
      for J in 1 .. JLast loop
        TranslatePi(C(Q.Sigma-J));
        Add(C(Q.Sigma-J),C(J));
        Mult(Half,C(J));
        Copy(C(J),C(Q.Sigma-J));
      end loop;
    elsif T=ASymm then
      SetZero(Q.A);
      if Q.Sigma=0 then SymmetrizePi(1,C(0),Check); end if;
      for J in 1 .. JLast loop
        TranslatePi(C(Q.Sigma-J));
        Sub(C(Q.Sigma-J),C(J));
        Mult(Half,C(J));
        Neg(C(J),C(Q.Sigma-J));
      end loop;
    end if;
  end Pi_Symmetrize;

  --- S=1:      [0-JMax]  1-JMax  2-JMax ... JMax-1  JMax  [JMax+1]
  --- S=0:  [-1-JMax]  0-JMax   1-JMax    ...    JMax-1  JMax  [JMax+1]

  procedure NegNabla(Q: in out FChain) is
    --- C(J) := C(J)-C(J+1)    nobla star
    C: CChain renames Q.C;
    C0: CosSin1 renames C(0);
  begin
    if Q.Sigma=0 then raise Parity_Error; end if;
    if STrunc then
      SetZero(C0.R,C0.P,C(JEMax));
      SetZero(C0.R,C0.P,C(-JMax));
    else
      Copy(C(-JMax),C(-JEMax));
    end if;
    for J in -JMax .. JMax loop Sub(C(J+1),C(J)); end loop;
    if Not_STrunc then
      Mult(One+InvRho2,C(JEMax));
      Mult(One+InvRho2,C(-JEMax));
    end if;
    Q.Sigma := 0;
    SetZero(Q.A);
  end NegNabla;

  procedure NegNabla(Q1: in FChain; Q2: in out FChain) is
  begin
    Copy(Q1,Q2);
    NegNabla(Q2);
  end NegNabla;

  procedure Nabla(Q: in out FChain) is
  begin
    Neg(Q);
    NegNabla(Q);
  end Nabla;

  procedure Nabla(Q1: in FChain; Q2: in out FChain) is
  begin
    Neg(Q1,Q2);
    NegNabla(Q2);
  end Nabla;

  --- S=0:  [-1-JMax]  0-JMax   1-JMax    ...    JMax-1  JMax  [JMax+1]
  --- S=1:      [0-JMax]  1-JMax  2-JMax ... JMax-1  JMax  [JMax+1]

  procedure NegNablaStar(Q: in out FChain) is
    --- C(J) := C(J)-C(J-1)    nobla
    Tmp: LP_Pointer;
    C: CChain renames Q.C;
  begin
    if Q.Sigma=1 then raise Parity_Error; end if;
    if STrunc then
      for J in reverse 1-JMax .. JMax loop Sub(C(J-1),C(J)); end loop;
    else
      PPool.Allocate(Tmp);
      declare
        P: Cossin1 renames Tmp.all.Data;
      begin
        Prod(One+Rho2,C(JEMax),P);           -- for J>JEMax
        AddProdToErr(One,C(JMax),C(JEMax));  -- for J=JEMax
        Union(P,C(JEMax));                   -- for J>=JEMax
        Prod(One+Rho2,C(-JEMax),P);
        AddProdToErr(One,C(-JMax),C(-JEMax));
        Union(P,C(-JEMax));                  -- belongs into C(-JMax)
      end;
      PPool.Recycle(Tmp);
      for J in reverse 1-JMax .. JMax loop Sub(C(J-1),C(J)); end loop;
      Copy(C(-JEMax),C(-JMax));
    end if;
    Q.Sigma := 1;
    SetZero(Q.A);
  end NegNablaStar;

  procedure NegNablaStar(Q1: in FChain; Q2: in out FChain) is
  begin
    Copy(Q1,Q2);
    NegNablaStar(Q2);
  end NegNablaStar;

  procedure NablaStar(Q: in out FChain) is
  begin
    Neg(Q);
    NegNablaStar(Q);
  end NablaStar;

  procedure NablaStar(Q1: in FChain; Q2: in out FChain) is
  begin
    Neg(Q1,Q2);
    NegNablaStar(Q2);
  end NablaStar;

  procedure Nabla1(Q: in out FChain) is
    --- Q.Sigma=0: C(J) := C(J)-C(J-1)      nobla
    --- Q.Sigma=1: C(J) := C(J+1)-C(J)      nabla
  begin
    if Q.Sigma=0 then NegNablaStar(Q); else Nabla(Q); end if;
  end Nabla1;

  procedure Nabla1(Q1: in FChain; Q2: in out FChain) is
    --- Q1.Sigma=0: C2(J) := C1(J)-C1(J-1)
    --- Q1.Sigma=1: C2(J) := C1(J+1)-C1(J)
  begin
    if Q1.Sigma=0 then NegNablaStar(Q1,Q2); else Nabla(Q1,Q2); end if;
  end Nabla1;

  procedure NablaStar0(Q: in out FChain) is
    --- Q.Sigma=0: C(J) := C(J-1)-C(J)          nabla star
    --- Q.Sigma=1: C(J) := C(J)-C(J+1)          nobla star
  begin
    if Q.Sigma=0 then NablaStar(Q); else NegNabla(Q); end if;
  end NablaStar0;

  procedure NablaStar0(Q1: in FChain; Q2: in out FChain) is
    --- Q1.Sigma=0: C2(J) := C1(J-1)-C1(J)
    --- Q1.Sigma=1: C2(J) := C1(J)-C1(J+1)
  begin
    if Q1.Sigma=0 then NablaStar(Q1,Q2); else NegNabla(Q1,Q2); end if;
  end NablaStar0;

  procedure Compose(C: in SPoly1; Q1: in FChain; Q2: in out FChain) is
    --- pointwise composition
    JEMin: constant Integer := Q1.Sigma-JEMax;
    C1: CChain renames Q1.C;
    C2: CChain renames Q2.C;
    Task_Error: Boolean := False;
    PC: Protected_Counter(JEMin);

    procedure Comp_Loop is
      J: Integer;
      Tmp: LP_Pointer;
    begin
      Proper_Rounding;
      PPool.Allocate(Tmp);
      loop
        PC.Next(J);
        exit when J>JEMax;
        if Not_STrunc or else (J>JEMin and then J<JEMax) then
          Compose_Part(C,C1(J),C2(J),Tmp.all.Data);
        end if;
      end loop;
      PPool.Recycle(Tmp);
    exception
      when others => Task_Error := True; raise;
    end Comp_Loop;

    NT: constant Natural := Reserve_Tasks(2*JMax,Comp_Parallel);
  begin
    Q2.Sigma := Q1.Sigma;
    if not IsZero(Q1.A) then raise Not_Implemented; end if;
    Copy(C(0),Q2.A);
    if NT=0 then
      Comp_Loop;
    else
      declare
        task type Comp_Task_Type is end Comp_Task_Type;
        task body Comp_Task_Type is begin Comp_Loop; end Comp_Task_Type;
        Comp_Task: array(1 .. NT) of Comp_Task_Type;
      begin
        Comp_Loop;
      end;
      Free_Tasks(NT);
      if Task_Error then raise Sorry with "Task_Error"; end if;
    end if;
  end Compose;

  procedure InvHelm(Omega,M: in Scalar; Q: in out FChain) is
    C: CChain renames Q.C;
    Omega2: Scalar;
  begin
    if not IsZero(Q.A) then
      Div(M,Q.A);
      Neg(Q.A);
    end if;
    Prod(Omega,Omega,Omega2);
    for J in Q.Sigma-JLast .. JLast loop InvHelm(Omega2,M,C(J)); end loop;
  end InvHelm;

  procedure InvHelm(Omega,M: in Scalar; Q1: in FChain; Q2: in out FChain) is
  begin
    Copy(Q1,Q2);
    InvHelm(Omega,M,Q2);
  end InvHelm;

  procedure Show1(N: in String; P: in FChain; Hide0: in Boolean := True) is
    use Flt_Vectors;
  begin
    if not IsZero(P) then
      Show1(N & "S ",P.Sigma);
      Show1(N & "A ",P.A);
      if STrunc then
        Show1(N,P.C(P.Sigma-JMax .. JMax),Hide0);
      else
        Show1(N,P.C,Hide0);
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Put(F: in File_Type; Q: in FChain; Decimal: in Boolean := False) is
    C: CChain renames Q.C;
    C0: CosSin1 renames C(0);
    Tmp: LP_Pointer;
  begin
    if not CheckRad(Q) then raise Inconsistent_Data; end if;
    Put(F,JEMax);
    Put(F,Q.Sigma);
    Put(F,Rho,Decimal);
    if AIO then Put(F,Q.A,Decimal); end if;
    for J in Q.Sigma-JMax .. JMax loop
      Put(F,C(J),Decimal);
    end loop;
    if STrunc then
      PPool.Allocate(Tmp);
      declare
        P: CosSin1 renames Tmp.all.Data;
      begin
        SetZero(C0.R,C0.P,P);
        Put(F,P,Decimal);
        Put(F,P,Decimal);
      end;
      PPool.Recycle(Tmp);
    else
      Put(F,C(JEMax),Decimal);
      Put(F,C(Q.Sigma-JEMax),Decimal);
    end if;
  end Put;

  procedure Get(F: in File_Type; Q: in out FChain; Decimal: in Boolean := False) is
    First: Boolean := True;
    Tail: Boolean := False;
    JEMaxSaved: Positive := 1;
    Tmp: LP_Pointer;
    R: Radius := Zero;
    Sigma: Parity := 0;
  begin
    Get(F,JEMaxSaved);
    Get(F,Sigma);
    Get(F,R,Decimal);
    if AIO then Get(F,Q.A,Decimal); else SetZero(Q.A); end if;
    PPool.Allocate(Tmp);
    declare
      P: CosSin1 renames Tmp.all.Data;
    begin
      for J in Sigma+1-JEMaxSaved .. JEMaxSaved-1 loop
        Get(F,P,Decimal);
        if First then SetZero(P.P,Sigma,Q); First := False; end if;
        Add_Site(J,P,Q);
      end loop;
      if Not_STrunc then
        Get(F,P,Decimal);
        if not IsZero(P) then
          Add_Tail(JEMaxSaved,P,Q);
          Tail := True;
        end if;
        Get(F,P,Decimal);
        if not IsZero(P) then
          Add_Tail(Q.Sigma-JEMaxSaved,P,Q);
          Tail := True;
        end if;
      end if;
    end;
    PPool.Recycle(Tmp);
    FixRad(Q);
    if R<Rho then
      Show1("had smaller rho:",R);
      if Tail then raise Domain_Error; end if;
    end if;
  end Get;

  procedure Write(FileName: in String; Q: in FChain; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,Q,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; Q: in out FChain; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,Q,Decimal);
    Close(F);
  end Read;

  procedure Phi2(Q: in FChain; Ph2: in out FChain; First: in Natural := 1) is
    --- Phi"(Nabla1 Q)
    --- First is the first nonzero coefficient of Phi' used
    CPhi2: constant SPoly1 := DerConvert(Phi1,First);
    Tmp: LF_Pointer;
  begin
    FPool.Allocate(Tmp);
    Nabla1(Q,Tmp.all.Data);
    Compose(CPhi2,Tmp.all.Data,Ph2);
    FPool.Recycle(Tmp);
  end Phi2;

  function Phi2(Q: FChain; First: Natural := 1) return FChain is
    Ph2: FChain;
  begin
    Phi2(Q,Ph2,First);
    return Ph2;
  end Phi2;

  procedure Psi2(Q: in FChain; Ps2: in out FChain; First: in Natural := 1) is
    --- Psi"(Q)
    --- First is the first nonzero coefficient of Psi' used
    CPsi2: constant SPoly1 := DerConvert(Psi1,First);
  begin
    Compose(CPsi2,Q,Ps2);
  end Psi2;

  function Psi2(Q: FChain; First: Natural := 1) return FChain is
    Ps2: FChain;
  begin
    Psi2(Q,Ps2,First);
    return Ps2;
  end Psi2;

  procedure Phi2Range(Q: in FChain; YMin,YMax: out Flt; Pow2: in Positive := 12) is
    --- upper and lower bounds on Phi"(Nabla1 Q)
    Tmp: LF_Pointer;
  begin
    FPool.Allocate(Tmp);
    declare
      Ph2: FChain renames Tmp.all.Data;
    begin
      Phi2(Q,Ph2,1);
      YRange(Ph2,YMin,YMax,Pow2);
    end;
    FPool.Recycle(Tmp);
  end Phi2Range;

  procedure Psi2Range(Q: in FChain; YMin,YMax: out Flt; Pow2: in Positive := 12) is
    --- upper and lower bounds on Phi"(Q)
    Tmp: LF_Pointer;
  begin
    FPool.Allocate(Tmp);
    declare
      Ps2: FChain renames Tmp.all.Data;
    begin
      Psi2(Q,Ps2,1);
      YRange(Ps2,YMin,YMax,Pow2);
    end;
    FPool.Recycle(Tmp);
  end Psi2Range;

  ------------------------ modes

  procedure Norm(M: in CMode; S: in out Scalar) is
  begin
    if M.Err then
      Copy(SOne,S);  --- error modes have norm 1
    else
      Prod(RadPower(M.I),RhoPower(abs(2*M.J-M.Sigma)),S);
    end if;
  end Norm;

  procedure MaxNorm(M: in CMode; S: out Radius) is
  begin
    if M.Err then
      S := One;      --- error modes have norm 1
    else
      S := RadPow(M.I)*RhoPow(abs(2*M.J-M.Sigma));
    end if;
  end MaxNorm;

  procedure Assign(M: in CMode; C: in out CSVector) is
    --- assuming that M is fine
    Par: Parity renames M.Par;
    I: EFreq renames M.I;
    J: JRangeE renames M.J;
    P: CosSin1 renames C(J);
  begin
    for K in C'Range loop SetZero(Rad,Par,C(K)); end loop;
    if M.Err then
      P.E(I) := InvRadPow(I)*InvRhoPow(abs(2*J-M.Sigma));
    else
      Copy(SOne,P.C(I));
    end if;
  end Assign;

  procedure Assign(M: in CMode; Q: in out FChain) is
  begin
    Q.Sigma := M.Sigma;
    SetZero(Q.A);
    Assign(M,Q.C);
  end Assign;

  procedure AddProd(M: in CMode; S: in Scalar; C: in out CSVector) is
    --- assuming M is fine
    J: JRangeE renames M.J;
    I: EFreq renames M.I;
    P: CosSin1 renames C(J);
  begin
    if not IsZero(S) then
      if M.Err then
        P.E(I) := P.E(I)+InvRadPow(I)*InvRhoPow(abs(2*J-M.Sigma))*MaxNorm(S);
      else
        Add(S,P.C(I));
      end if;
    end if;
  end AddProd;

  procedure AddProd(M: in CMode; S: in Scalar; Q: in out FChain; Check: in Boolean) is
    --- assuming M is fine
  begin
    if Check and then (M.Sigma /= Q.Sigma) then raise Parity_Error; end if;
    AddProd(M,S,Q.C);
  end AddProd;

  procedure Extract_Coeff(M: in CMode; C: in out CSVector; S: in out Scalar) is
    --- assuming M.Err=False
    --- assuming M and C are of the same type
    --- S encloses the coefficient
    I: EFreq renames M.I;
    P: CosSin1 renames C(M.J);
    PC: FPoly1 renames P.C;
    PE: Errs1 renames P.E;
  begin
    if M.Simple then                       -- move a coeff (into S)
      SetZero(S);
      Swap(PC(I),S);
    else                                   -- move a coeff (into S) and shift errors
      if I=1 and then (PE(0) /= Zero) then -- cleanup
        PE(1) := PE(1)+PE(0)/Rad;
        PE(0) := Zero;
      end if;
      if PE(I)=Zero then                   -- no need to shift errors
        SetZero(S);
        Swap(PC(I),S);
      else
        BallAt0(PE(I),S);
        Add(PC(I),S);
        SetZero(PC(I));
        PE(I+1) := PE(I+1)+PE(I)/Rad;
        PE(I) := Zero;
      end if;
    end if;
  end Extract_Coeff;

  procedure Extract_CosSin1_Error(M: in CMode; C: in out CSVector; S: in out Scalar) is
    --- assuming M.Err=True and M.Sigma-JMax <= J <= JMax
    --- assuming M and C are of the same type
    --- S enloses the norm
    J: JRangeE renames M.J;
    I: EFreq renames M.I;
    P: CosSin1 renames C(J);
    PC: FPoly1 renames P.C;
    PE: Errs1 renames P.E;
    SC,SE: Radius := Zero;
  begin
    if M.Simple then
      BallAt0(PE(I),S);
      PE(I) := Zero;
    else
      if I=1 and then (PE(0) /= Zero) then     -- cleanup
        PE(1) := PE(1)+PE(0)/Rad;
        PE(0) := Zero;
      end if;
      for K in reverse I .. EDeg loop -- move CosSin error with index >= I (into S)
        SE := Rad*SE+PE(K);
        PE(K) := Zero;
      end loop;
      for K in reverse I .. FDeg loop
        SC := Rad*SC;
        if not IsZero(PC(K)) then
          SC := SC+MaxNorm(PC(K));
          SetZero(PC(K));
        end if;
      end loop;
      BallAt0(RadPow(I)*RhoPow(abs(2*J-M.Sigma))*(SC+SE),S);
    end if;
  end Extract_CosSin1_Error;

  procedure Extract_FChain_Error(M: in CMode; C: in out CSVector; S: in out Scalar) is
    --- assuming M.Err=True, and J=JEMax or J=M.Sigma-JEMax
    --- assuming M and C are of the same type
    --- S encloses the norm
    I: Parity renames M.Par;
    J: JRangeE renames M.J;
    P: CosSin1 renames C(J);
    PE: Errs1 renames P.E;
    SE: Radius := Zero;
  begin
    if M.I /= I then raise Inconsistent_Data; end if;
    if M.Simple then
      BallAt0(RadPow(I)*PE(I),S);  --- new
      PE(I) := Zero;
    else
      for K in reverse 0 .. EDeg loop -- move CosSin1 error with index >= 0 (into S)
        SE := Rad*SE+PE(K);
        PE(K) := Zero;
      end loop;
      BallAt0(RhoPow(2*JEMax-M.Sigma)*SE,S);
    end if;
  end Extract_FChain_Error;

  procedure Extract(M: in CMode; C: in out CSVector; S: in out Scalar) is
    JEMin: constant Integer := M.Sigma-JEMax;
    --- assuming M is consistent
  begin
    if (not M.Err) then
      Extract_Coeff(M,C,S);
    elsif (M.J<JEMax) and then (M.J>JEMin) then
      Extract_CosSin1_Error(M,C,S);
    else                             --- J=JEMax or J=JEMin (last)
      Extract_FChain_Error(M,C,S);
    end if;
  end Extract;

  procedure Extract(M: in CMode; Q: in out FChain; S: in out Scalar; Check: in Boolean) is
    pragma Unreferenced(Check);
    JEMin: constant Integer := M.Sigma-JEMax;
    --- assuming M is consistent
  begin
    Extract(M,Q.C,S);
    if (M.Sigma/=Q.Sigma) then raise Parity_Error; end if;
    if (M.J=JEMin) and then (not IsZero(Q)) then raise Sorry; end if; --- check for partition
  end Extract;

  procedure Check(M: in CMode) is
    JMin: constant Integer := M.Sigma-JMax;
    Bad: exception;
  begin
    if STrunc and then (M.Err or (not M.Simple)) then raise Bad; end if;
    if (not M.Err) and then ((M.I>FDeg) or (M.J>JMax) or (M.J<JMin)) then raise Bad; end if;
  exception
    when Bad =>
      Show0("!!! inconsistent mode:");
      Show(M);
      raise Inconsistent_Data;
  end Check;

  procedure ShowIJ(N: in String; M: in CMode; NewLine: in Boolean := True) is
  begin
    if M.Err then
      Show0(N & " Err:  J=" & Strng(M.J,3) & " I=" & Strng(M.I,3),NewLine);
    else
      Show0(N & " Coe:  J=" & Strng(M.J,3) & " I=" & Strng(M.I,3),NewLine);
    end if;
  end ShowIJ;

  procedure ShowType(M: in CMode) is
  begin
    Show1("   Par =",M.Par);
    Show1("     S =",M.Sigma);
    Show1("Simple =",M.Simple);
  end ShowType;

  procedure Show(M: in CMode) is
  begin
    ShowType(M);
    Show1("   Err =",M.Err);
    Show1("     J =",M.J);
    Show1("     I =",M.I);
  end Show;

  function SameType(M1,M2: CMode) return Boolean is
  begin
    return (M1.Par=M2.Par) and (M1.Sigma=M2.Sigma) and (M1.Simple=M2.Simple);
  end SameType;

  procedure Show(M: in CModes) is
    F: constant Positive := M'First;
    MF: CMode renames M(F);
    OK: Boolean;
  begin
    if F /= 1 then raise Inconsistent_Data; end if;
    ShowType(MF);
    ShowIJ("   1",MF);
    for K in 2 .. M'Last loop
      OK := SameType(MF,M(K));
      ShowIJ(Strng(K,4),M(K),OK);
      if not OK then Show0("  !!! wrong type"); end if;
    end loop;
  end Show;

  function Compare(Sigma: Parity; J1,J2: Integer) return Integer is
    --- order for S=0 is: 0,1,Sigma-1,2,Sigma-2,...
    --- order for S=1 is:   1,Sigma-1,2,Sigma-2,...
    N1: constant Natural := IMax(J1,Sigma-J1);
    N2: constant Natural := IMax(J2,Sigma-J2);
  begin
    if N1>N2 then return  1; end if;
    if N1<N2 then return -1; end if;
    if J1<J2 then return  1; end if;
    if J1>J2 then return -1; end if;
    return 0;
  end Compare;

  procedure Check_Adjacent(M1,M2: in CMode) is
    Cmp: constant Integer := Compare(M1.Sigma,M1.J,M2.J);
    Bad: exception;
  begin
    if not SameType(M1,M2) then raise Bad; end if;
    if Cmp>0 then raise Bad; end if;
    if Cmp=0 then
      if (M2.I <= M1.I) then raise Bad; end if;
      if M1.Err then raise Bad; end if;
    else --- Cmp<0
         --- could also check value of M2.I
      if not (STrunc or M1.Err) then raise Bad; end if;
    end if;
  exception
    when Bad =>
      Show0("!!! inconsistent mode pair:");
      Show(M1);
      Show(M2);
      raise Inconsistent_Data;
  end Check_Adjacent;

  procedure Check(M: in CModes) is
    --- does not check partition property, but see Extract
  begin
    if M'First /= 1 then raise Inconsistent_Data; end if;
    Check(M(1));
    for K in 2 .. M'Last loop
      Check(M(K));
      Check_Adjacent(M(K-1),M(K));
    end loop;
  end Check;

  function Extract(J: Integer; M: CModes) return CModes is
    --- extract modes M(K) with M(K).J=J
    L: Natural := 0;
    MP: CModes := M;
  begin
    for K in M'Range loop
      if M(K).J=J then L := L+1; MP(L) := M(K); end if;
    end loop;
    return MP(1 .. L);
  end Extract;

  function Find(M: CModes; Par: Parity; I: Natural; J: Integer) return Natural is
  begin
    for K in M'Range loop
      if M(K).J=J and then M(K).I=I and then M(K).Par=Par then return K; end if;
    end loop;
    return 0;
  end Find;

  function Gen_Make(JCut: Positive; Par,Sigma: Parity; Simple,Include_Errs: Boolean) return CModes is
    Include_CS_Errs: Boolean renames Include_Errs;
    Include_Ch_Errs: constant Boolean := Include_Errs and (JCut=JMax);

    procedure Build(M: out CModes; Count: out Natural) is
      --- order for Sigma=0 is: 0,1,Sigma-1,2,Sigma-2,...
      --- order for Sigma=1 is:   1,Sigma-1,2,Sigma-2,...
      procedure Set_Mode(J: Integer; I: in Natural; Err: in Boolean) is
      begin
        Count := Count+1;
        if Count <= M'Last then M(Count) := (I,J,Par,Sigma,Err,Simple); end if;
      end Set_Mode;
      J,D: Integer;
    begin
      Count := 0;
      for N in Sigma .. JCut loop
        D := MaxDeg(N);
        J := N;
        loop
          for I in Par .. D loop
            Set_Mode(J,I,False);            --- coefficient
          end loop;
          if Include_CS_Errs then
            Set_Mode(J,IMax(D+1,Par),True); --- CosSin1 error
          end if;
          J := Sigma-J;
          exit when J=N;
        end loop;
      end loop;
      if Include_Ch_Errs then
        Set_Mode(JEMax,Par,True);            --- FChain error
        Set_Mode(Sigma-JEMax,Par,True);      --- FChain error
      end if;
    end Build;
    Count: Natural;
  begin
    declare
      M: CModes(1..1);
    begin
      Build(M,Count);
    end;
    declare
      M: CModes(1..Count);
    begin
      Build(M,Count);
      return M;
    end;
  end Gen_Make;

  function Make(Siz: Positive; Par,Sigma: Parity; Simple,Trunc: Boolean) return CModes is
    Simp: constant Boolean := Simple or Trunc or STrunc;
    Include_Errs: constant Boolean := not (Trunc or STrunc);
    function MaxDeg(N: Natural) return Integer is begin return IMin(Siz-N,FDeg); end MaxDeg;
    function GMake is new Gen_Make (MaxDeg => MaxDeg);
  begin
    return GMake(JMax,Par,Sigma,Simp,Include_Errs);
  end Make;

  ---------------------

  procedure CenterRange(Q: in FChain; J1,J2: out PlotRange; Small: in Radius) is
    C: CChain renames Q.C;
    VMax: Radius := Zero;
    V: Flt_Vectors.Vector(PlotRange);
  begin
    if Small>One then raise Sorry; end if;
    for J in PLotRange loop
      V(J) := MaxNorm(C(J));
      VMax := RMax(V(J),VMax);
    end loop;
    J1 := -JMax;
    while V(J1)<Small*VMax loop J1 := J1+1; end loop;
    J2 := JMax;
    while V(J2)<Small*VMax loop J2 := J2-1; end loop;
  end CenterRange;

  procedure PutPlot(F: in File_Type; Q: in FChain; T: in Flt; J1,J2: in PlotRange; PutT,PutJ: in Boolean := False) is
    --- Q.C(J1 .. J2) at time T
    use Int_IO,FLt_IO;
    JF: constant Integer := IMin(J1,J2);
    JL: constant Integer := IMax(J1,J2);
    Digs: constant Positive := 4;
    S: Flt;
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    declare
      P: CosSin1 renames Tmp.all.Data;
    begin
      for J in JF .. JL loop
        Component(J,Q,P);
        S := NumVal(P,T)+Approx(Q.A);
        if PutT then Put(F,T,3,Digs,4); end if;
        if PutJ then Put(F,J,5); end if;
        Put(F,S,3,Digs,4);
        New_Line(F);
      end loop;
    end;
    PPool.Recycle(Tmp);
  end PutPlot;

  procedure PutPlot(F: in File_Type; Q: in FChain; Steps: in Positive; J1,J2: in PlotRange; PutT,PutJ,Blank: in Boolean := False) is
    --- Q.C(J1 .. J2) at Steps different times
    Pi2: constant Flt := 6.28318530718;
    T: Flt;
  begin
    for N in 0 .. Steps loop
      T := Pi2*Flt(N)/Flt(Steps);
      PutPlot(F,Q,T,J1,J2,PutT,PutJ);
      if Blank then New_Line(F); end if;
    end loop;
  end PutPlot;

  procedure WritePlot(FileName: in String; Q: in FChain; Steps: in Positive) is
    --- includes Steps, -JMax, JMax
    use Int_IO;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,Steps,5);
    New_Line(F);
    Put(F,-JMax,5);
    New_Line(F);
    Put(F,JMax,5);
    New_Line(F);
    PutPlot(F,Q,Steps,-JMax,JMax);
    Close(F);
  end WritePlot;

  procedure WritePlot3d(FileName: in String; Q: in FChain; Steps: in Positive; Small: in Radius) is
    J1,J2: Integer;
    F: File_Type;
  begin
    CenterRange(Q,J1,J2,Small);
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    PutPlot(F,Q,Steps,J1,J2,True,True,True);
    Close(F);
  end WritePlot3d;

end CosSins1.Chain;
