with Strings;
use Strings;

pragma Elaborate_All (Strings);

package body CosSins1 is

  use SV;

  procedure Set_CosSins1_Default(Rad: in Radius; Par: in Parity) is
  begin
    Default_Radius := Rad;
    Default_Parity := Par;
  end Set_CosSins1_Default;

  function Set_CosSins1_Default(Rad: Radius; Par: Parity) return Boolean is
  begin
    Default_Radius := Rad;
    Default_Parity := Par;
    return True;
  end Set_CosSins1_Default;

  function Info(Dummy: CosSin1) return Scalar_Info is
    pragma Unreferenced(Dummy);
  begin
    return PInfo;
  end Info;

  function Is_Zero(P: Parity; C: FPoly1) return Boolean is
  begin
    for I in P .. FDeg loop
      if not IsZero(C(I)) then return False; end if;
    end loop;
    return True;
  end Is_Zero;

  function Is_Zero(E: Errs1) return Boolean is
  begin
    if Not_STrunc then
      for I in EFreq loop
        if E(I) /= Zero then return False; end if;
      end loop;
    end if;
    return True;
  end Is_Zero;

  procedure LiftE0(P: in out CosSin1) is
    --- use only if Not_STrunc
    E: Errs1 renames P.E;
  begin
    if (E(0) /= Zero) then
      E(1) := E(1)+E(0)/P.R;
      E(0) := Zero;
    end if;
  end LiftE0;

  ----------------------------------

  function IsZero(P: CosSin1) return Boolean is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop
      if not IsZero(C(I)) then return False; end if;
    end loop;
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsZero;

  function IsSharp(P: CosSin1) return Boolean is
  begin
    if Not_STrunc then
      declare
        C: FPoly1 renames P.C;
        E: Errs1 renames P.E;
      begin
        for I in P.P .. FDeg loop
          if not IsSharp(C(I)) then return False; end if;
        end loop;
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsSharp;

  function IsPoly(P: CosSin1) return Boolean is
  begin
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop
          if E(I) /= Zero then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end IsPoly;

  function "="(P1,P2: CosSin1) return Boolean is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.R /= P2.R then return False; end if;
    if P1.P /= P2.P then return False; end if;
    for I in P1.P .. FDeg loop
      if C1(I) /= C2(I) then return False; end if;
    end loop;
    if not STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop
          if E1(I) /= E2(I) then return False; end if;
        end loop;
      end;
    end if;
    return True;
  end "=";

  procedure FillZero(P: in out CosSin1) is
  begin
    SetZero(P.C);
    if Not_STrunc then P.E := (others => Zero); end if;
  end FillZero;

  procedure SetZero(PR: in Radius; PP: in Parity; P: in out CosSin1) is
  begin
    P.R := PR;
    P.P := PP;
    SetZero(P.C);
    if Not_STrunc then P.E := (others => Zero); end if;
  end SetZero;

  procedure SetZero(P: in out CosSin1) is
  begin
    P.R := Default_Radius;
    P.P := Default_Parity;
    SetZero(P.C);
    if Not_STrunc then P.E := (others => Zero); end if;
  end SetZero;

  procedure Copy(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Copy(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end Copy;

  procedure Swap(P1,P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    R: Flt;
  begin
    if P1.P /= P2.P then raise Not_Implemented; end if;
    R := P1.R; P1.R := P2.R; P2.R := R;
    for I in P1.P .. FDeg loop Swap(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop
          R := E1(I); E1(I) := E2(I); E2(I) := R;
        end loop;
      end;
    end if;
  end Swap;

  ----------------------------------

  function Center0(P: CosSin1) return Boolean is
    --- symmetric about zero
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop
      if not Center0(C(I)) then return False; end if;
    end loop;
    return True;
  end Center0;

  function Contains0(P: CosSin1) return Logical is
    --- very crude
    C: FPoly1 renames P.C;
    L: Logical := True;
    LI: Logical;
  begin
    for I in P.P .. FDeg loop
      LI := Contains0(C(I));
      if LI=False then return False; end if;
      if LI=Uncertain then L := Uncertain; end if;
    end loop;
    return L;
  end Contains0;

  procedure BallAt0(R,PR: in Flt; PP: in Parity; P: in out CosSin1) is
  begin
    SetZero(PR,PP,P);
    if Not_STrunc then
      P.E(0) := R;
      if P.P=1 then LiftE0(P); end if;
    end if;
  end BallAt0;

  procedure BallAt0(R: in Flt; P: in out CosSin1) is
  begin
    BallAt0(R,Default_Radius,Default_Parity,P);
  end BallAt0;

  function BallAt0(R: Flt) return CosSin1 is
    P: CosSin1;
  begin
    BallAt0(R,P);
    return P;
  end BallAt0;

  procedure ToErr(P: in out CosSin1) is
    --- zero-centered set containing P
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop ToErr(C(I)); end loop;
  end ToErr;

  procedure ToErr(P1: in CosSin1; P2: in out CosSin1) is
    --- zero-centered set containing P1
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop ToErr(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end ToErr;

  function ToErr(P: CosSin1) return CosSin1 is
    --- zero-centered set containing P
    Q: CosSin1;
  begin
    ToErr(P,Q);
    return Q;
  end ToErr;

  procedure Center(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    if Not_STrunc then
      for I in P.P .. FDeg loop Center(C(I)); end loop;
      P.E := (others => Zero);
    end if;
  end Center;

  procedure Center(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Center(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := (others => Zero); end if;
  end Center;

  function Center(P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Center(P,Q);
    return Q;
  end Center;

  procedure ModCenter(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop ModCenter(C(I)); end loop;
  end ModCenter;

  procedure ModCenter(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop ModCenter(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end ModCenter;

  function ModCenter(P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    ModCenter(P,Q);
    return Q;
  end ModCenter;

  procedure SplitErr(P: in CosSin1; PC,PE: in out CosSin1) is
    --- split P into a sharp PC and an error PE
  begin
    Center(P,PC);
    ModCenter(P,PE);
  end SplitErr;

  procedure ErrMult(R: in Radius; P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    if Not_STrunc then
      for I in P.P .. FDeg loop ErrMult(R,C(I)); end loop;
      declare
        E: Errs1 renames P.E;
      begin
        for I in EFreq loop E(I) := R*E(I); end loop;
      end;
    end if;
  end ErrMult;

  procedure Union(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    if P2.R>P1.R then P2.R := P1.R; end if;
    for I in P1.P .. FDeg loop Union(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop E2(I) := RMax(E1(I),E2(I)); end loop;
      end;
    end if;
  end Union;

  function Union(P1,P2: CosSin1) return CosSin1 is
    Q: CosSin1 := P2;
  begin
    Union(P1,Q);
    return Q;
  end Union;

  ------------------------------------------------------------

  procedure Neg(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop Neg(C(I)); end loop;
  end Neg;

  procedure Neg(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Neg(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end Neg;

  function "-"(P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Neg(P,Q);
    return Q;
  end "-";

  procedure Add(I: in Integer; P: in out CosSin1) is
  begin
    if P.P=1 then raise Parity_Error; end if;
    Add(I,P.C(0));
  end Add;

  procedure Add(S: in Scalar; P: in out CosSin1) is
  begin
    if P.P=1 then raise Parity_Error; end if;
    Add(S,P.C(0));
  end Add;

  procedure Add(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    if P2.R>P1.R then P2.R := P1.R; end if;
    for I in P1.P .. FDeg loop Add(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop E2(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Add;

  procedure Sum(P1,P2: in CosSin1; P3: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    C3: FPoly1 renames P3.C;
  begin
    P3.R := RMin(P1.R,P2.R);
    if P1.P /= P2.P then raise Parity_Error; end if;
    P3.P := P1.P;
    for I in P1.P .. FDeg loop Sum(C1(I),C2(I),C3(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
        E3: Errs1 renames P3.E;
      begin
        for I in EFreq loop E3(I) := E1(I)+E2(I); end loop;
      end;
    end if;
  end Sum;

  function "+"(P1,P2: CosSin1) return CosSin1 is
    P3: CosSin1;
  begin
    Sum(P1,P2,P3);
    return P3;
  end "+";

  procedure Sub(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    if P2.R>P1.R then P2.R := P1.R; end if;
    for I in P1.P .. FDeg loop Sub(C1(I),C2(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
      begin
        for I in EFreq loop E2(I) := E2(I)+E1(I); end loop;
      end;
    end if;
  end Sub;

  procedure Diff(P1,P2: in CosSin1; P3: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    C3: FPoly1 renames P3.C;
  begin
    P3.R := RMin(P1.R,P2.R);
    if P1.P /= P2.P then raise Parity_Error; end if;
    P3.P := P1.P;
    for I in P1.P .. FDeg loop Diff(C1(I),C2(I),C3(I)); end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
        E3: Errs1 renames P3.E;
      begin
        for I in EFreq loop E3(I) := E1(I)+E2(I); end loop;
      end;
    end if;
  end Diff;

  function "-"(P1,P2: CosSin1) return CosSin1 is
    P3: CosSin1;
  begin
    Diff(P1,P2,P3);
    return P3;
  end "-";

  procedure Mult(R: in Flt; P: in out CosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar;
  begin
    if R=Zero then
      FillZero(P);
    else
      for I in P.P .. FDeg loop Mult(R,C(I),Tmp); end loop;
      if Not_STrunc then
        declare
          E: Errs1 renames P.E;
          A: constant Flt := abs(R);
        begin
          for I in EFreq loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(R: in Flt; P1: in CosSin1; P2: in out CosSin1) is
  begin
    if R=One then
      Copy(P1,P2);
    elsif R=NegOne then
      Neg(P1,P2);
    else
      P2.R := P1.R;
      P2.P := P1.P;
      if R=Zero then
        FillZero(P2);
      else
        declare
          C1: FPoly1 renames P1.C;
          C2: FPoly1 renames P2.C;
        begin
          for I in P1.P .. FDeg loop Prod(R,C1(I),C2(I)); end loop;
        end;
        if Not_STrunc then
          declare
            E1: Errs1 renames P1.E;
            E2: Errs1 renames P2.E;
            A: constant Flt := abs(R);
          begin
            for I in EFreq loop E2(I) := A*E1(I); end loop;
          end;
        end if;
      end if;
    end if;
  end Prod;

  function "*"(R: Flt; P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Prod(R,P,Q);
    return Q;
  end "*";

  procedure AddProd(R: in Flt; P1: in CosSin1; P2: in out CosSin1) is
  begin
    if R=Zero then
      null;
    elsif R=One then
      Add(P1,P2);
    elsif R=NegOne then
      Sub(P1,P2);
    else
      if P1.P /= P2.P then raise Parity_Error; end if;
      if P2.R>P1.R then P2.R := P1.R; end if;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        Tmp: Scalar;
      begin
        for I in P1.P .. QuasiDeg1(C1) loop AddProd(R,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
          A: constant Flt := abs(R);
        begin
          for I in EFreq loop E2(I) := E2(I)+A*E1(I); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(Q: in Rational; P: in out CosSin1) is
    QNum: constant LInt := LNum(Q);
    C: FPoly1 renames P.C;
  begin
    if QNum=0 then
      FillZero(P);
    else
      for I in P.P .. QuasiDeg1(C) loop Mult(Q,C(I)); end loop;
      if not IsPoly(P) then
        declare
          R: constant Flt := Flt(abs(QNum))/Flt(LDen(Q));
          E: Errs1 renames P.E;
        begin
          for I in EFreq loop E(I) := R*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Mult(S: in Scalar; P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    if IsZero(S) then
      FillZero(P);
    else
      for I in P.P .. FDeg loop Mult(S,C(I)); end loop;
      if Not_STrunc then
        declare
          A: constant Flt := MaxNorm(S);
          E: Errs1 renames P.E;
        begin
          for I in EFreq loop E(I) := A*E(I); end loop;
        end;
      end if;
    end if;
  end Mult;

  procedure Prod(S: in Scalar; P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    if IsZero(S) then
      FillZero(P2);
    else
      for I in P1.P .. FDeg loop Prod(S,C1(I),C2(I)); end loop;
      if Not_STrunc then
        declare
          A: constant Flt := MaxNorm(S);
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
        begin
          for I in EFreq loop E2(I) := A*E1(I); end loop;
        end;
      end if;
    end if;
  end Prod;

  function "*"(S: Scalar; P: CosSin1) return CosSin1 is
    Q: CosSin1;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";

  procedure AddProd(S: in Scalar; P1: in CosSin1; P2: in out CosSin1) is
  begin
    if P1.P /= P2.P then raise Parity_Error; end if;
    if not IsZero(S) then
      if P2.R>P1.R then P2.R := P1.R; end if;
      declare
        C1: FPoly1 renames P1.C;
        C2: FPoly1 renames P2.C;
        Tmp: Scalar;
      begin
        for I in P1.P .. QuasiDeg1(C1) loop AddProd(S,C1(I),C2(I),Tmp); end loop;
      end;
      if Not_STrunc then
        declare
          E1: Errs1 renames P1.E;
          E2: Errs1 renames P2.E;
          A: constant Flt := MaxNorm(S);
        begin
          for I in EFreq loop E2(I) := E2(I)+A*E1(I); end loop;
        end;
      end if;
    end if;
  end AddProd;

  procedure Mult(P1: in CosSin1; P2,Tmp: in out CosSin1) is
  begin
    Prod(P1,P2,Tmp);
    Copy(Tmp,P2);
  end Mult;

  procedure Mult(P1: in CosSin1; P2: in out CosSin1) is
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    Prod(P1,P2,Tmp.all.Data);
    Copy(Tmp.all.Data,P2);
    PPool.Recycle(Tmp);
  end Mult;

  --  2*cos(m)*cos(n) =  cos(m-n) + cos(m+n)
  --  2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
  --  2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
  --  2*sin(m)*cos(n) =  sin(m-n) + sin(m+n)

  procedure CCProd2(P1P,P2P: in Parity; C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CosSin1) is
    --- add twice C1*C2
    --- does not set P3.P or P3.R
    --- for efficiency, check first whether C1=0 or C2=0
    C3: FPoly1 renames P3.C;
    E3: Errs1 renames P3.E;
    K: Integer;
    Tmp: Scalar;
  begin
    if P1P=0 then
      if P2P=0 then                           --- 2*cos(m)*cos(n) = cos(m-n) + cos(m+n)
        for M in 0 .. FDeg loop
          if A1(M) /= Zero then
            for N in 0 .. FDeg loop
              if A2(N) /= Zero then
                K := M-N;                               --- freq diff
                AddProd(C1(M),C2(N),C3(abs(K)),Tmp);
                K := M+N;                               --- freq sum
                if K <= FDeg then
                  AddProd(C1(M),C2(N),C3(K),Tmp);
                elsif Not_STrunc then
                  E3(K) := E3(K)+A1(M)*A2(N);
                end if;
              end if;
            end loop;
          end if;
        end loop;
      else                                 --- 2*cos(m)*sin(n) = -sin(m-n) + sin(m+n)
        for M in 0 .. FDeg loop
          if A1(M) /= Zero then
            for N in 1 .. FDeg loop
              if A2(N) /= Zero then
                K := M-N;                            --- freq diff
                if K>0 then
                  SubProd(C1(M),C2(N),C3(K),Tmp);    ---    M>N
                elsif K<0 then
                  AddProd(C1(M),C2(N),C3(-K),Tmp);   ---    M<N
                end if;
                K := M+N;                            --- freq sum
                if K <= FDeg then
                  AddProd(C1(M),C2(N),C3(K),Tmp);
                elsif Not_STrunc then
                  E3(K) := E3(K)+A1(M)*A2(N);
                end if;
              end if;
            end loop;
          end if;
        end loop;
      end if;
    elsif P2P=0 then                    --- 2*sin(m)*cos(n) = sin(m-n) + sin(m+n)
      for M in 1 .. FDeg loop
        if A1(M) /= Zero then
          for N in 0 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                            --- freq diff
              if K>0 then
                AddProd(C1(M),C2(N),C3(K),Tmp);    ---    M>N
              elsif K<0 then
                SubProd(C1(M),C2(N),C3(-K),Tmp);   ---    M<N
              end if;
              K := M+N;                            --- freq sum
              if K <= FDeg then
                AddProd(C1(M),C2(N),C3(K),Tmp);
              elsif Not_STrunc then
                E3(K) := E3(K)+A1(M)*A2(N);
              end if;
            end if;
          end loop;
        end if;
      end loop;
    else                                 --- 2*sin(m)*sin(n) =  cos(m-n) - cos(m+n)
      for M in 1 .. FDeg loop
        if A1(M) /= Zero then
          for N in 1 .. FDeg loop
            if A2(N) /= Zero then
              K := M-N;                             --- freq diff
              AddProd(C1(M),C2(N),C3(abs(K)),Tmp);
              K := M+N;                             --- freq sum
              if K <= FDeg then
                SubProd(C1(M),C2(N),C3(K),Tmp);
              elsif Not_STrunc then
                E3(K) := E3(K)+A1(M)*A2(N);
              end if;
            end if;
          end loop;
        end if;
      end loop;
    end if;
  end CCProd2;

  procedure CEProd2(P1P: in Parity; A1: in CAbs; E2: in Errs1; E3: in out Errs1; W: in Weights) is
    --- add twice A1*E2
    --- for efficiency, check first whether A1=0 or E2=0
    E3F: Radius renames E3(0);
    E3L: Radius renames E3(EDeg);
    K: Integer;
  begin
    for M in P1P .. FDeg loop
      if A1(M) /= Zero then
        for N in 0 .. EDeg loop
          if E2(N) /= Zero then
            K := N-M;                          --- freq diff
            if K<0 then                        ---   M>N
              E3F := E3F+W(-K)*A1(M)*E2(N);
            else                               ---   M<N
              E3(K) := E3(K)+A1(M)*E2(N);
            end if;
            K := N+M;                          --- freq sum
            if K <= EDeg then
              E3(K) := E3(K)+A1(M)*E2(N);
            else
              E3L := E3L+W(K-EDeg)*A1(M)*E2(N);
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end CEProd2;

  procedure EEProd2A(E1,E2: in Errs1; E3: in out Errs1; W: in Weights) is
    --- add twice E1*E2
    --- slower version, more accurate
    --- for efficiency, check first whether E1=0 or E2=0
    E3F: Radius renames E3(0);
    E3L: Radius renames E3(EDeg);
    K: Integer;
  begin
    for M in 0 .. EDeg loop
      if E1(M) /= Zero then
        for N in 0 .. EDeg loop
          if E2(N) /= Zero then
            K := N-M;                         --- freq diff
            E3F := E3F+W(abs(K))*E1(M)*E2(N);
            K := N+M;                         --- freq sum
            if K <= EDeg then
              E3(K) := E3(K)+E1(M)*E2(N);
            else
              E3L := E3L+W(K-EDeg)*E1(M)*E2(N);
            end if;
          end if;
        end loop;
      end if;
    end loop;
  end EEProd2A;

  procedure EEProd2B(E1,E2: in Errs1; E3: in out Errs1; W: in Weights) is
    --- add twice E1*E2
    --- faster version, less accurate
    --- for efficiency, check first whether E1=0 or E2=0
    R: Radius renames W(1);
    ESum1,ESum2: Flt := Zero;
  begin
    for K in reverse EFreq loop
      ESum1 := R*ESum1+E1(K);
      ESum2 := R*ESum2+E2(K);
    end loop;
    E3(0) := Two*(ESum1+ESum2);
  end EEProd2B;

  procedure Make_NormFactors(R: in Radius; W: out Weights) is
    --- upper bounds (not too precise)
  begin
    if R<One then raise Undefined; end if;
    W(0) := One;
    for I in 1 .. EDeg loop
      W(I) := W(I-1)*R;
    end loop;
  end Make_NormFactors;

  protected body Protected_NormFactors is
    procedure Get(R: in Radius; W: out Weights) is
    begin
      if R /= RSaved then
        Make_NormFactors(R,WSaved);
      end if;
      W := WSaved;
    end Get;
  end Protected_NormFactors;

  procedure Prod(P1,P2: in CosSin1; P3: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    E1: Errs1 renames P1.E;
    E2: Errs1 renames P2.E;
    E3: Errs1 renames P3.E;
    ZC1: constant Boolean := Is_Zero(P1.P,C1);
    ZE1: constant Boolean := Is_Zero(E1);
    ZC2,ZE2: Boolean;
    A1,A2: CAbs;
  begin
    FillZero(P3);
    P3.R := RMin(P1.R,P2.R);
    P3.P := (P1.P+P2.P) mod 2;
    if ZC1 and then ZE1 then return; end if; -- P1=0
    ZC2 := Is_Zero(P2.P,C2);
    ZE2 := Is_Zero(E2);
    if ZC2 then
      if ZE2 then return; end if;            -- P2=0
    else
      for I in P2.P .. FDeg loop A2(I) := MaxNorm(C2(I)); end loop;
    end if;
    if not ZC1 then
      for I in P1.P .. FDeg loop A1(I) := MaxNorm(C1(I)); end loop;
      if not ZC2 then
        CCProd2(P1.P,P2.P,C1,C2,A1,A2,P3);
      end if;
    end if;
    if not (ZE1 and ZE2) then
      declare
        W: Weights;
      begin
        NormFactors.Get(P3.R,W);
        if not (ZC1 or ZE2) then
          CEProd2(P1.P,A1,E2,E3,W);
        end if;
        if not (ZC2 or ZE1) then
          CEProd2(P2.P,A2,E1,E3,W);
        end if;
        if not (ZE1 or ZE2) then
          EEProd2(E1,E2,E3,W);
        end if;
        if P3.P=1 then LiftE0(P3); end if;
      end;
    end if;
    Mult(SHalf,P3);
  end Prod;

  function "*"(P1,P2: CosSin1) return CosSin1 is
    P3: CosSin1;
  begin
    Prod(P1,P2,P3);
    return P3;
  end "*";

  procedure AddProd(P1,P2: in CosSin1; P3: in out CosSin1) is
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    Prod(P1,P2,Tmp.all.Data);
    Add(Tmp.all.Data,P3);
    PPool.Recycle(Tmp);
  end AddProd;

  procedure AddProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1) is
  begin
    Prod(P1,P2,Tmp);
    Add(Tmp,P3);
  end AddProd;

  procedure SubProd(P1,P2: in CosSin1; P3: in out CosSin1) is
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    Prod(P1,P2,Tmp.all.Data);
    Sub(Tmp.all.Data,P3);
    PPool.Recycle(Tmp);
  end SubProd;

  procedure SubProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1) is
  begin
    Prod(P1,P2,Tmp);
    Sub(Tmp,P3);
  end SubProd;

  procedure Div(R: in Flt; P: in out CosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar;
  begin
    for I in P.P .. FDeg loop Div(R,C(I),Tmp); end loop;
    if Not_STrunc then
      declare
        E: Errs1 renames P.E;
        A: constant Flt := abs(R);
      begin
        for I in EFreq loop E(I) := E(I)/A; end loop;
      end;
    end if;
  end Div;

  procedure Quot(P1: in CosSin1; R: in Flt; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop
      Quot(C1(I),R,C2(I));
    end loop;
    if Not_STrunc then
      declare
        E1: Errs1 renames P1.E;
        E2: Errs1 renames P2.E;
        A: constant Flt := abs(R);
      begin
        for I in EFreq loop E2(I) := E1(I)/A; end loop;
      end;
    end if;
  end Quot;

  function "/"(P: CosSin1; R: Flt) return CosSin1 is
    Q: CosSin1;
  begin
    Quot(P,R,Q);
    return Q;
  end "/";

  ------------------------------------------------

  function Get_Precision(P: CosSin1) return Positive is
  begin
    return Get_Precision(P.C(1));
  end Get_Precision;

  procedure Assign(I: in Integer; P: in out CosSin1) is
  begin
    SetZero(Default_Radius,0,P);
    Assign(I,P.C(0));
  end Assign;

  procedure Assign(Q: in Rational; P: in out CosSin1) is
  begin
    SetZero(Default_Radius,0,P);
    Assign(Q,P.C(0));
  end Assign;

  procedure Assign(R: in Flt; P: in out CosSin1) is
  begin
    SetZero(Default_Radius,0,P);
    Assign(R,P.C(0));
  end Assign;

  procedure Assign(S: in Scalar; P: in out CosSin1) is
  begin
    SetZero(Default_Radius,0,P);
    Copy(S,P.C(0));
  end Assign;

  procedure Show1(N: in String; P: in CosSin1; Hide0: in Boolean := True) is
    use Flt_Vectors;
  begin
    if not IsZero(P) then
      Show1(N & "Rad ",P.R);
      if P.P=0 then
        Show1(N & "Cos ",P.C(P.P .. FDeg));
      else
        Show1(N & "Sin ",P.C(P.P .. FDeg));
      end if;
      if not IsPoly(P) then
        Show1(N & "Err ",P.E);
      end if;
    elsif not Hide0 then
      Show0(N & "0");
    end if;
  end Show1;

  procedure Show2(N: in String; P1,P2: in CosSin1; Hide0: in Boolean := True) is
    use Flt_Vectors;
  begin
    if P1.P /= P2.P then raise Not_Implemented; end if;
    if not (IsZero(P1) and IsZero(P2)) then
      Show2(N & "Rad ",P1.R,P2.R);
      if P1.P=0 then
        Show2(N & "Cos ",P1.C(P1.P .. FDeg),P2.C(P2.P .. FDeg));
      else
        Show2(N & "Sin ",P1.C(P1.P .. FDeg),P2.C(P2.P .. FDeg));
      end if;
      if not (IsPoly(P1) and IsPoly(P2)) then
        Show2(N & "Err ",P1.E,P2.E);
      end if;
    elsif not Hide0 then
      Show0(N & "0 0");
    end if;
  end Show2;

  function QuasiDeg(P: Parity; C: FPoly1) return Natural is
  begin
    for I in reverse P .. FDeg loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return 0;
  end QuasiDeg;

  function QuasiDeg1(C: FPoly1) return Natural is
  begin
    for I in reverse 2 .. FDeg loop
      if not IsZero(C(I)) then return I; end if;
    end loop;
    return 1;
  end QuasiDeg1;

  procedure Put(F: in File_Type; P: in CosSin1; Decimal: in Boolean := False) is
    use Flt_Vectors;
    C: FPoly1 renames P.C;
    E: Errs1 renames P.E;
    DC: constant Integer := QuasiDeg(P.P,C);
    DE: Integer;
  begin
    Put(F,P.R,Decimal);
    Put(F,P.P);
    Put(F,DC);
    for I in P.P .. DC loop
      Put(F,C(I),Decimal);
    end loop;
    if STrunc then
      Put(F,0);
      Put(F,Zero,Decimal);
    else
      DE := QuasiDeg(E);
      Put(F,DE);
      for I in 0 .. DE loop
        Put(F,E(I),Decimal);
      end loop;
    end if;
  end Put;

  procedure Get(F: in File_Type; P: in out CosSin1; Decimal: in Boolean := False) is
    DC,DE: Integer := 0;
    R: Flt := Zero;
    S: Scalar;
  begin
    FillZero(P);
    Get(F,P.R,Decimal);
    Get(F,P.P);
    Get(F,DC);
    for I in P.P .. DC loop
      Get(F,S,Decimal);
      AddCoeff(I,S,P);
    end loop;
    Get(F,DE);
    for I in 0 .. DE loop
      Get(F,R,Decimal);
      AddErr(I,R,P);
    end loop;
    if Not_STrunc and then P.P=1 then LiftE0(P); end if;
  end Get;

  procedure Write(FileName: in String; P: in CosSin1; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out CosSin1; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

  -----------------------------------------

  function IsReal(P: CosSin1) return Boolean is
    C: FPoly1 renames P.C;
  begin
    if PInfo.IsComplex then
      for I in P.P .. FDeg loop
        if not IsReal(C(I)) then return False; end if;
      end loop;
      if not IsPoly(P) then raise Not_Certain; end if;
    end if;
    return True;
  end IsReal;

  procedure Real_Part(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    if PInfo.IsComplex then
      for I in P.P .. FDeg loop Real_Part(C(I)); end loop;
    end if;
  end Real_Part;

  procedure Imag_Part(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    for I in P.P .. FDeg loop Imag_Part(C(I)); end loop;
    if not (STrunc or else PInfo.IsComplex) then
      P.E := (others => Zero);
    end if;
  end Imag_Part;

  procedure Adjoint(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    if PInfo.IsComplex then
      for I in P.P .. FDeg loop Adjoint(C(I)); end loop;
    end if;
  end Adjoint;

  procedure Adjoint(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop Adjoint(C1(I),C2(I)); end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end Adjoint;

  procedure Simple_Random(P: in out CosSin1) is
    --- assumes P.P and P.R have been set
    R: constant Flt := P.R*Flt(9)/Flt(8); --- ad hoc
    C: FPoly1 renames P.C;
    W: Weights;
  begin
    NormFactors.Get(R,W);
    for I in FFreq loop
      Simple_Random(C(I));
      Div(W(I),C(I));
    end loop;
    if Not_STrunc then P.E := (others => Zero); end if;
  end Simple_Random;

  procedure Simple_Random(PR: in Radius; PP: in Parity; P: in out CosSin1) is
  begin
    P.R := PR;
    P.P := PP;
    Simple_Random(P);
  end Simple_Random;

  function MaxNorm(P: CosSin1) return Radius is
    R: Radius renames P.R;
    C: FPoly1 renames P.C;
    N: Flt := Zero;
  begin
    for I in reverse 1 .. FDeg loop
      N := R*(N+MaxNorm(C(I)));
    end loop;
    if P.P=0 then N := N+MaxNorm(C(0)); end if;
    if Not_STrunc then
      if R>P.R then raise Domain_Error; end if;
      declare
        E: Errs1 renames P.E;
        ESum: Flt := Zero;
      begin
        for I in reverse EFreq loop
          ESum := R*ESum+E(I);
        end loop;
        N := N+ESum;
      end;
    end if;
    return N;
  end MaxNorm;

  procedure Norm(P: in CosSin1; S: in out Scalar) is
    R: Radius renames P.R;
    C: FPoly1 renames P.C;
    Tmp: Scalar;
  begin
    SetZero(S);
    for I in reverse 1 .. FDeg loop
      Norm(C(I),Tmp);
      Add(Tmp,S);
      Mult(R,S,Tmp);
    end loop;
    if P.P=0 then
      Norm(C(0),Tmp);
      Add(Tmp,S);
    end if;
    if not IsPoly(P) then
      declare
        E: Errs1 renames P.E;
        ESum: Flt := Zero;
      begin
        for I in reverse EFreq loop
          ESum := R*ESum+E(I);
        end loop;
        BallAt0(ESum,Tmp);
      end;
      Add(Tmp,S);
    end if;
  end Norm;

  function Norm(P: CosSin1) return Scalar is
    S: Scalar;
  begin
    Norm(P,S);
    return S;
  end Norm;

  procedure YRange(P: in CosSin1; YInf,YSup: out Flt; Pow2: in Positive := 12) is
    Steps: constant Integer := 2**Pow2;
    D: constant Natural := QuasiDeg(P.P,P.C);
    Pi2: constant Flt := Flt(210828715)/Flt(33554432); --- bit larger than 2*Pi
    Eps: constant Flt := Pi2/Flt(Steps);
    DX: constant Scalar := Scal(Eps);
    X,Y: Scalar;
  begin
    ValAt0(P,Y,D,True);
    YInf := Inf(Y);
    YSup := Sup(Y);
    if D>0 then
      for N in 1 .. Steps-1 loop
        Prod(Flt(N),DX,X);       --- X := Pi2*N/Steps
        Val(P,X,Y,D,True);
        YInf := RMin(Inf(Y),YInf);
        YSup := RMax(Sup(Y),YSup);
      end loop;
    end if;
    if Not_Strunc then
      declare
        C: FPoly1 renames P.C;
        E: Errs1 renames P.E;
        MaxDer,R: Radius := Zero;
      begin
        for I in reverse 1 .. D loop
          MaxDer := MaxDer+Flt(I)*MaxNorm(C(I));
        end loop;
        for I in reverse EFreq loop R := R+E(I); end loop;
        R := R+MaxDer*Half*Eps;
        YInf := -(R-YInf);
        YSup := YSup+R;
      end;
    end if;
  end YRange;

  function Inf(P: CosSin1; Pow2: Positive := 12) return Flt is
    YInf,YSup: Flt;
  begin
    YRange(P,YInf,YSup,Pow2);
    return YInf;
  end Inf;

  function Sup(P: CosSin1; Pow2: Positive := 12) return Flt is
    YInf,YSup: Flt;
  begin
    YRange(P,YInf,YSup,Pow2);
    return YSup;
  end Sup;

  function LInftyNorm(P: CosSin1; Pow2: Positive := 12) return Radius is
    YInf,YSup: Flt;
  begin
    YRange(P,YInf,YSup,Pow2);
    return RMax(abs(YInf),abs(YSup));
  end LInftyNorm;

  procedure Average(P: in CosSin1; S: in out Scalar) is
  begin
    if P.P=1 then
      SetZero(S);
    elsif (STrunc or else P.E(0)=Zero) then
      Copy(P.C(0),S);
    else
      BallAt0(P.E(0),S);
      Add(P.C(0),S);
    end if;
  end Average;

  function Average(P: CosSin1) return Scalar is
    S: Scalar;
  begin
    Average(P,S);
    return S;
  end Average;

  procedure SetAverage0(P: in out CosSin1) is
  begin
    if P.P=0 then SetZero(P.C(0)); end if;
    if Not_STrunc then LiftE0(P); end if;
  end SetAverage0;

  procedure SplitOffAverage(P: in out CosSin1; S: in out Scalar) is
    C: Scalar renames P.C(0);
    E: Errs1 renames P.E;
  begin
    if P.P=1 then
      SetZero(S);
      if Not_STrunc then
        E(1) := E(1)+E(0)/P.R;
        E(0) := Zero;
      end if;
    else
      if STrunc or else E(0)=Zero then
        Copy(C,S);
      else
        BallAt0(E(0),S);
        Add(C,S);
        E(1) := E(1)+E(0)/P.R;
        E(0) := Zero;
      end if;
      SetZero(C);
    end if;
  end SplitOffAverage;

  procedure Coeff(I: in Natural; P: in CosSin1; S: in out Scalar) is
  begin
    if I=0 then
      Average(P,S);
    elsif STrunc then
      if I>FDeg then
        SetZero(S);
      else
        Copy(P.C(I),S);
      end if;
    else
      declare
        R: Radius renames P.R;
        E: Errs1 renames P.E;
        ESum: Flt := E(0);
      begin
        if I>EDeg then
          for K in 1 .. EDeg loop ESum := E(K)+ESum/R; end loop;
          BallAt0(NPower(One/R,I-EDeg)*ESum,S);
        else
          for K in 1 .. I loop ESum := E(K)+ESum/R; end loop;
          BallAt0(ESum,S);
          Add(P.C(I),S);
        end if;
      end;
    end if;
  end Coeff;

  function Coeff(I: Natural; P: CosSin1) return Scalar is
    S: Scalar;
  begin
    Coeff(I,P,S);
    return S;
  end Coeff;

  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out CosSin1) is
  begin
    if not IsZero(S) then
      if I <= FDeg then
        if I<P.P then raise Undefined; end if;
        Add(S,P.C(I));
      elsif Not_STrunc then
        AddErr(I,MaxNorm(S),P);
      end if;
    end if;
  end AddCoeff;

  procedure AddProdToErr(R: in Flt; P1: in CosSin1; P2: in out CosSin1) is
    --- add R*P1 to error P2.E
    C1: FPoly1 renames P1.C;
    E1: Errs1 renames P1.E;
    E2: Errs1 renames P2.E;
    A: constant Flt := abs(R);
  begin
    if not (STrunc or else (R=Zero) or else IsZero(P1)) then
      if P1.P /= P2.P then raise Parity_Error; end if;
      if P2.R>P1.R then P2.R := P1.R; end if;
      for I in P1.P .. FDeg loop
        if not IsZero(C1(I)) then
          E2(I) := E2(I)+A*MaxNorm(C1(I));
        end if;
      end loop;
      for I in EFreq loop E2(I) := E2(I)+A*E1(I); end loop;
    end if;
  end AddProdToErr;

  function EffErrTerm(I: Natural; P: CosSin1) return Radius is
    --- if I=1 and P.P=1, then include P.E(0) with P.E(1)
    --- I>EDeg causes index error
  begin
    if STrunc then return Zero; end if;
    if I>P.P then return P.E(I); end if;
    if P.P=0 then return P.E(0); end if;
    if I=0 then return Zero; end if;
    return P.E(1)+P.E(0)/P.R;
  end EffErrTerm;

  procedure AddErr(I: in Natural; R: in Flt; P: in out CosSin1) is
    E: Errs1 renames P.E;
  begin
    if STrunc then
      null;
    elsif I <= EDeg then
      E(I) := E(I)+abs(R);
    else
      E(EDeg) := E(EDeg)+NPower(P.R,I-EDeg)*abs(R);
    end if;
  end AddErr;

  procedure AddBall(R: in Radius; P: in out CosSin1) is
  begin
    if not STrunc then
      P.E(0) := P.E(0)+R;
      if P.P=1 then LiftE0(P); end if;
    end if;
  end AddBall;

  procedure ValAt0(P: in CosSin1; S: in out Scalar; D: in Natural; Is_Poly: in Boolean) is
    --- D is assumed to be QuasiDeg(P.P,P.C);
    C: FPoly1 renames P.C;
    ESum: Flt;
  begin
    if P.P=0 then
      if Is_Poly then
        SetZero(S);
      else
        ESum := Zero;
        for I in reverse EFreq loop
          ESum := ESum+EffErrTerm(I,P);
        end loop;
        BallAt0(ESum,S);
      end if;
      for I in reverse 0 .. D loop
        Add(C(I),S);
      end loop;
    else
      SetZero(S);
    end if;
  end ValAt0;

  procedure ValAt0(P: in CosSin1; S: in out Scalar) is
  begin
    ValAt0(P,S,QuasiDeg(P.P,P.C),IsPoly(P));
  end ValAt0;

  --- cos((n+1).)=cos*cos(n.)-sin*sin(n.)
  --- sin((n+1).)=sin*cos(n.)+cos*sin(n.)
  --- need to transpose

  procedure Val(P: in CosSin1; S1: in Scalar; S2: in out Scalar; D: in Natural; Is_Poly: in Boolean) is
    --- D is assumed to be QuasiDeg(P.P,P.C);
    C: FPoly1 renames P.C;
    ESum: Flt;
  begin
    if IsZero(S1) then ValAt0(P,S2,D,Is_Poly); return; end if;
    if Is_Poly then
      if D=0 then
        if P.P=0 then Copy(C(0),S2); else SetZero(S2); end if;
        return;
      end if;
      SetZero(S2);
    else
      if not IsReal(S1) then raise Not_Implemented; end if;
      ESum := Zero;
      for I in reverse EFreq loop
        ESum := ESum+EffErrTerm(I,P);
      end loop;
      BallAt0(ESum,S2);
    end if;
    declare
      Cos1: constant Scalar := Cos(S1);
      Sin1: constant Scalar := Sin(S1);
      CC,SS,CTmp,STmp,Tmp: Scalar;
    begin
      if P.P=0 then
        Copy(C(D),CC);
        SetZero(SS);
      else
        SetZero(CC);
        Copy(C(D),SS);
      end if;
      for I in reverse 0 .. D-1 loop
        Prod(Cos1,CC,CTmp);
        Prod(Sin1,SS,Tmp);
        Add(Tmp,CTmp);
        Prod(Sin1,CC,Tmp);
        Prod(Cos1,SS,STmp);
        Sub(Tmp,STmp);
        if P.P=0 then
          Sum(CTmp,C(I),CC);
          Copy(STmp,SS);
        else
          Copy(CTmp,CC);
          exit when I=0;
          Sum(STmp,C(I),SS);
        end if;
      end loop;
      Add(CC,S2);
    end;
  end Val;

  procedure Val(P: in CosSin1; S1: in Scalar; S2: in out Scalar) is
  begin
    Val(P,S1,S2,QuasiDeg(P.P,P.C),IsPoly(P));
  end Val;

  function Val(P: CosSin1; S: Scalar) return Scalar is
    V: Scalar;
  begin
    Val(P,S,V);
    return V;
  end Val;

  procedure TranslatePi(P: in out CosSin1) is
    --- C(odd) := -C(odd)
    C: FPoly1 renames P.C;
    I: Integer := 1;
  begin
    while I <= FDeg loop Neg(C(I)); I := I+2; end loop;
  end TranslatePi;

  procedure TranslatePi(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    Odd: Boolean := (P1.P=1);
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop
      if Odd then Neg(C1(I),C2(I)); else Copy(C1(I),C2(I)); end if;
      Odd := not Odd;
    end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end TranslatePi;

  procedure NegTranslatePi(P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
    Odd: Boolean := (P1.P=1);
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    for I in P1.P .. FDeg loop
      if Odd then Copy(C1(I),C2(I)); else Neg(C1(I),C2(I)); end if;
      Odd := not Odd;
    end loop;
    if Not_STrunc then P2.E := P1.E; end if;
  end NegTranslatePi;

  procedure SymmetrizePi(SP: in Parity; P: in out CosSin1; Check: in Boolean := False) is
    --- SP=0: symmetric, C(odd) := 0
    --- SP=1: antisymmetric, C(even) := 0
    C: FPoly1 renames P.C;
    E: Errs1 renames P.E;
    I: Integer := 1-SP;
  begin
    while I <= FDeg loop
      if not IsZero(C(I)) then
        if Check then raise Sorry; end if;
        SetZero(C(I));
      end if;
      I := I+2;
    end loop;
    if Not_STrunc then
      if P.P=1 then LiftE0(P); end if;
      I := 1-SP;
      while I<EDeg loop
        E(I+1) := E(I+1)+E(I);
        E(I) := Zero;
        I := I+2;
      end loop;
    end if;
  end SymmetrizePi;

  function Convert(C: Flt_Poly1; First: Natural := 0) return SPoly1 is
    --- first cut C(I) for I<First
    B: SPoly1(C'Range);
  begin
    for I in C'Range loop
      if I<First then
        SetZero(B(I));
        --        if C(I) /= Zero then Show1("cut ",C(I)); end if;
      else
        Assign(C(I),B(I));
      end if;
    end loop;
    return B;
  end Convert;

  procedure Compose_Part(C: in SPoly1; P1: in CosSin1; P2,Tmp: in out CosSin1) is
    --- Compose without C(0)
    D: constant Integer := QuasiDeg(C);
  begin
    if D=0 then
      SetZero(P1.R,0,P2);
    else
      Prod(C(D),P1,P2);
      for I in reverse 1 .. D-1 loop
        AddCoeff(0,C(I),P2); --- fails if P2.P=1 and C(I)/=0
        Mult(P1,P2,Tmp);
      end loop;
    end if;
  end Compose_Part;

  procedure Compose(C: in SPoly1; P1: in CosSin1; P2: in out CosSin1) is
    Tmp: LP_Pointer;
  begin
    PPool.Allocate(Tmp);
    Compose_Part(C,P1,P2,Tmp.all.Data);
    PPool.Recycle(Tmp);
    AddCoeff(0,C(0),P2);
  end Compose;

  function DerConvert(C: Flt_Poly1; First: Natural := 0) return SPoly1 is
    --- first cut C(I) for I<First, then differentiate
    use Flt_Vectors;
    D: constant Natural := EffDeg(C);
  begin
    if D=0 then return SPoly1'(0 => SZero); end if;
    declare
      DC: SPoly1(0 .. D-1);
    begin
      for I in 1 .. D loop
        if I<First then
          --          if C(I) /= Zero then Show1("cut ",C(I)); end if;
          SetZero(DC(I-1));
        else
          Assign(C(I),DC(I-1));
          Mult(Flt(I),DC(I-1));
        end if;
      end loop;
      return DC;
    end;
  end DerConvert;

  function Der(C: SPoly1) return SPoly1 is
    D: constant Natural := EffDeg(C);
  begin
    if D=0 then return SPoly1'(0 => SZero); end if;
    declare
      DC: SPoly1(0 .. D-1);
    begin
      Copy(C(1),DC(0));
      for N in 2 .. D loop
        Prod(Flt(N),C(N),DC(N-1));
      end loop;
      return DC;
    end;
  end Der;

  procedure Make_DerFactors(R1,R2: in Radius; Fac: out Weights) is
    Exp1: constant Flt := Flt(410105312)/Flt(150869313); -- lower bound
    N: Integer;
    F: Flt;
    S,T: Scalar;
  begin
    if R2>=R1 then raise Undefined; end if;
    Assign(R1,T);
    Quot(T,R2,S);     -- R1/R2 > 1
    Log(S,T);
    Inv(T,S);         -- 1/Log(R1/R2)
    F := MaxNorm(S);
    if F<Flt(EDeg) then
      N :=  Natural(Flt'Ceiling(F));
    else
      N := EDeg;
    end if;
    F := F/Exp1;      -- 1/[Log(R1/R2)*Exp(1)]
    if N >= EDeg then
      Fac := (others => F);
    else
      for I in 0 .. N-1 loop Fac(I) := F; end loop;
      F := R2/R1;
      Fac(N) := NPower(F,N);
      for I in N+1 .. EDeg loop
        Fac(I) := F*Fac(I-1);
      end loop;
      for I in N .. EDeg loop
        Fac(I) := Flt(I)*Fac(I);
      end loop;
    end if;
  end Make_DerFactors;

  protected body Protected_DerFactors is
    procedure Get(R1,R2: in Radius; Fac: out Weights) is
    begin
      if (R1 /= RSaved1) or else (R2 /= RSaved2) then
        Make_DerFactors(R1,R2,FacSaved);
      end if;
      Fac := FacSaved;
    end Get;
  end Protected_DerFactors;

  procedure Der(P: in out CosSin1; R: in Radius) is
    C: FPoly1 renames P.C;
    Tmp: Scalar renames C(0);
  begin
    if P.P=0 then
      P.P := 1;
      Neg(C(1));
      for I in 2 .. QuasiDeg1(C) loop
        Mult(Flt(-I),C(I),Tmp);
      end loop;
    else
      P.P := 0;
      SetZero(C(0));
      for I in 2 .. QuasiDeg1(C) loop
        Mult(Flt(I),C(I),Tmp);
      end loop;
    end if;
    if not IsPoly(P) then
      declare
        E: Errs1 renames P.E;
        Fac: Weights;
      begin
        DerFactors.Get(P.R,R,Fac);
        for I in EFreq loop
          E(I) := Fac(I)*E(I);
        end loop;
      end;
    end if;
  end Der;

  procedure Der(P1: in CosSin1; P2: in out CosSin1; R: in Radius) is
  begin
    Copy(P1,P2);
    Der(P2,R);
  end Der;

  procedure AntiDer(P: in out CosSin1) is
    C: FPoly1 renames P.C;
    Tmp: Scalar renames C(0);
  begin
    if P.P=0 then
      P.P := 1;
      if not IsZero(C(0)) then raise Undefined; end if;
      for I in 2 .. QuasiDeg1(C) loop
        Div(Flt(I),C(I),Tmp);
      end loop;
    else
      P.P := 0;
      Neg(C(1));
      for I in 2 .. QuasiDeg1(C) loop
        Div(Flt(-I),C(I),Tmp);
      end loop;
    end if;
    if not IsPoly(P) then
      declare
        E: Errs1 renames P.E;
      begin
        LiftE0(P);
        for I in 2 .. EDeg loop
          E(I) := E(I)/Flt(I);
        end loop;
      end;
    end if;
  end AntiDer;

  procedure AntiDer(P1: in CosSin1; P2: in out CosSin1) is
  begin
    Copy(P1,P2);
    AntiDer(P2);
  end AntiDer;

  function Integral(P: CosSin1; S1,S2: Scalar) return Scalar is
    Tmp: LP_Pointer;
    V1,V2,A: Scalar;
  begin
    PPool.Allocate(Tmp);
    Copy(P,Tmp.all.Data);
    SetZero(Tmp.all.Data.C(0));
    AntiDer(Tmp.all.Data);
    Val(Tmp.all.Data,S1,V1);
    Val(Tmp.all.Data,S2,V2);
    PPool.Recycle(Tmp);
    Sub(V1,V2);
    Average(P,A);
    if not IsZero(A) then
      Diff(S2,S1,V1);
      Mult(A,V1);
      Add(V1,V2);
    end if;
    return V2;
  end Integral;

  function InfAbs(S: Scalar) return Radius is
    R: Flt := Inf(S);
  begin
    if R >= Zero then return R; end if;
    R := -Sup(S);
    if R >= Zero then return R; end if;
    return Zero;
  end InfAbs;

  procedure InvHelm(Omega2,M: in Scalar; P: in out CosSin1) is
    --- inverse of Helmholtz operator Omega2*D^2-M
    C: FPoly1 renames P.C;
    E: Errs1 renames P.E;
    S: Scalar;
  begin
    if P.P=0 and then (not IsZero(C(0))) then
      Neg(M,S);
      Div(S,C(0));
    end if;
    for I in 1 .. FDeg loop
      if not IsZero(C(I)) then
        Prod(Flt(-(I*I)),Omega2,S);
        Sub(M,S);
        Div(S,C(I));
      end if;
    end loop;
    if not IsPoly(P) then
      if P.P=1 then LiftE0(P); end if;
      for I in P.P .. EDeg loop
        if I=EDeg or (E(I) /= Zero) then
          Prod(Flt(I*I),Omega2,S);
          Add(M,S);
          E(I) := E(I)/InfAbs(S);
        end if;
      end loop;
      if Inf(S)<Zero then raise Sorry; end if; --- last denominator has to be positive
    end if;
  end InvHelm;

  procedure InvHelm(Omega2,M: in Scalar; P1: in CosSin1; P2: in out CosSin1) is
    --- inverse of Helmholtz operator Omega2*D^2-M
  begin
    Copy(P1,P2);
    InvHelm(Omega2,M,P2);
  end InvHelm;

  function PowersOf(S: Scalar; MaxPower: Positive) return SVector is
    V: SVector(0 .. MaxPower);
  begin
    Copy(SOne,V(0));
    Copy(S,V(1));
    for N in 2 .. MaxPower loop
      Prod(V(N/2),V(N-N/2),V(N));
    end loop;
    return V;
  end PowersOf;

  -----------------------------

  procedure NumTrunc(Siz: in Natural; P1: in CosSin1; P2: in out CosSin1) is
    C1: FPoly1 renames P1.C;
    C2: FPoly1 renames P2.C;
  begin
    P2.R := P1.R;
    P2.P := P1.P;
    if Siz >= FDeg then
      for I in P1.P .. FDeg loop Center(C1(I),C2(I)); end loop;
    else
      for I in P1.P .. Siz loop Center(C1(I),C2(I)); end loop;
      for I in Siz+1 .. FDeg loop SetZero(C2(I)); end loop;
    end if;
    if Not_STrunc then P2.E := (others => Zero); end if;
  end NumTrunc;

  function NumVal(P: CosSin1; T: Flt) return Flt is
    use Flt_EF;
    C: FPoly1 renames P.C;
    D: constant Integer := QuasiDeg(P.P,C);
    S: Flt := Zero;
  begin
    if not IsReal(P) then raise Not_Implemented; end if;
    if P.P=1 then
      if T /= Zero then
        for I in reverse 1 .. D loop S := S+Approx(C(I))*Sin(Flt(I)*T); end loop;
      end if;
    elsif T=Zero then
      for I in reverse 0 .. D loop S := S+Approx(C(I)); end loop;
    else
      for I in reverse 0 .. D loop S := S+Approx(C(I))*Cos(Flt(I)*T); end loop;
    end if;
    return S;
  end NumVal;

  procedure PutPlot(F: in File_Type; P: in CosSin1; Steps: in Positive := 64; PutX: in Boolean := False) is
    use FLt_IO;
    Digs: constant Positive := 4;
    Pi2: constant Flt := 6.28318530718;
    X,Y: Flt;
  begin
    for N in 0 .. Steps loop
      X := Pi2*Flt(N)/Flt(Steps);
      Y := NumVal(P,X);
      if PutX then Put(F,X,3,Digs,4); end if;
      Put(F,Y,3,Digs,4);
      New_Line(F);
    end loop;
  end PutPlot;

  function GuessRho(P: CosSin1) return Radius is
    DMin: constant Integer := FDeg/5;
    DMax: constant Integer := (4*FDeg)/5;
    C: FPoly1 renames P.C;
    N: Integer := 0;
    A: Flt;
    B: Flt := MaxNorm(C(DMin));
    Q: Flt := One;
    use Flt_EF;
  begin
    for I in DMin+1 .. DMax loop
      A := B;
      B := MaxNorm(C(I));
      if (A /= Zero) and then (B /= Zero) then
        Q := Q*A/B;
        N := N+1;
      end if;
    end loop;
    return Exp(Log(Q)/Flt(N));
  end GuessRho;

  procedure EvenOdd(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    for K in 0 .. FDeg loop
      case (K mod 4) is
        when 0 => SetZero(C(K));
        when 2 => SetZero(C(K));
        when 3 => Neg(C(K));
        when others => null;
      end case;
    end loop;
    P.P := 1;
  end EvenOdd;

  procedure OddEven(P: in out CosSin1) is
    C: FPoly1 renames P.C;
  begin
    for K in 0 .. FDeg loop
      case (K mod 4) is
        when 0 => SetZero(C(K));
        when 2 => SetZero(C(K));
        when 3 => Neg(C(K));
        when others => null;
      end case;
    end loop;
    P.P := 0;
  end OddEven;

end CosSins1;
