with Ada.Text_IO, Scalars, Logicals, Globals, Ints, Rationals, Flts, Flt_Vectors;
use Ada.Text_IO, Scalars, Logicals, Globals, Ints, Rationals, Flts;
with Recycling, Vectors;

pragma Elaborate_All (Ada.Text_IO,Scalars,Logicals,Globals,Ints,Rationals,Flt_Vectors);
pragma Elaborate_All (Recycling,Vectors);

generic

  FDeg: in Positive;  --- max degree

  -------------------- begin_include scalars Scalar

  type Scalar is private;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsNumbers(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>; -- S1 constains S2
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with function DiskAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure CenterDisk(S: in out Scalar) is <>;
  with procedure CenterDisk(S1: in Scalar; S2: in out Scalar) is <>;
  with function CenterDisk(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ModNumbers(S: in out Scalar) is <>;
  with procedure ModNumbers(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModNumbers(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  -------------------- end_include scalars

  with package SV is new Vectors (Component => Scalar);

package CosSins1 is      -- cos series and sin series in 1 variable

  EDeg: constant Positive := 2*FDeg;
  subtype FFreq is Integer range 0 .. FDeg;
  subtype EFreq is Integer range 0 .. EDeg;
  subtype SPoly1 is SV.Polynom1;
  subtype SVector is SV.Vector;
  subtype FPoly1 is SVector(FFreq);
  subtype FltVec is Flt_Vectors.Vector;
  subtype Flt_Poly1 is Flt_Vectors.Polynom1;
  subtype Errs1 is Flt_Poly1(EFreq);

  type CosSin1 is
    record
      R: Radius;  -- domain of analyticity includes |Im z|<Log(R)
      P: Parity;  -- 0: even  1: odd
      C: FPoly1;  -- cos(K.) if P=0,  sin(K.) if P=1
      E: Errs1;   -- higher order frequency errors (require same norm factor as coeffs)
    end record;

  --- try to avoid relying on defaults, if possible
  procedure Set_CosSins1_Default(Rad: in Radius; Par: in Parity);
  function Set_CosSins1_Default(Rad: Radius; Par: Parity) return Boolean;

  --- basic
  function Info(Dummy: CosSin1) return Scalar_Info;
  function IsZero(P: CosSin1) return Boolean;
  function IsSharp(P: CosSin1) return Boolean;          --- polynomial with sharp coeffs
  function IsPoly(P: CosSin1) return Boolean;
  function "="(P1,P2: CosSin1) return Boolean;
  procedure FillZero(P: in out CosSin1);                --- sets neither radius not parity
  procedure SetZero(PR: in Radius; PP: in Parity; P: in out CosSin1);
  procedure SetZero(P: in out CosSin1);                 --- sets default radius and parity
  procedure Copy(P1: in CosSin1; P2: in out CosSin1);
  procedure Swap(P1,P2: in out CosSin1);

  --- sets
  function Center0(P: CosSin1) return Boolean;
  function Contains0(P: CosSin1) return Logical;
  procedure BallAt0(R,PR: in Flt; PP: in Parity; P: in out CosSin1);
  procedure BallAt0(R: in Flt; P: in out CosSin1);      --- sets default radius and parity
  function BallAt0(R: Flt) return CosSin1;              --- sets default radius and parity
  procedure ToErr(P: in out CosSin1);
  procedure ToErr(P1: in CosSin1; P2: in out CosSin1);
  function ToErr(P: CosSin1) return CosSin1;
  procedure Center(P: in out CosSin1);
  procedure Center(P1: in CosSin1; P2: in out CosSin1);
  function Center(P: CosSin1) return CosSin1;
  procedure ModCenter(P: in out CosSin1);
  procedure ModCenter(P1: in CosSin1; P2: in out CosSin1);
  function ModCenter(P: CosSin1) return CosSin1;            --- zero center
  procedure SplitErr(P: in CosSin1; PC,PE: in out CosSin1);
  procedure ErrMult(R: in Radius; P: in out CosSin1);
  procedure Union(P1: in CosSin1; P2: in out CosSin1);
  function Union(P1,P2: CosSin1) return CosSin1;

  --- basic arithmetic
  procedure Neg(P: in out CosSin1);
  procedure Neg(P1: in CosSin1; P2: in out CosSin1);
  function "-"(P: CosSin1) return CosSin1;
  procedure Add(I: in Integer; P: in out CosSin1);
  procedure Add(S: in Scalar; P: in out CosSin1);
  procedure Add(P1: in CosSin1; P2: in out CosSin1);
  procedure Sum(P1,P2: in CosSin1; P3: in out CosSin1);
  function "+"(P1,P2: CosSin1) return CosSin1;
  procedure Sub(P1: in CosSin1; P2: in out CosSin1);
  procedure Diff(P1,P2: in CosSin1; P3: in out CosSin1);
  function "-"(P1,P2: CosSin1) return CosSin1;
  procedure Mult(R: in Flt; P: in out CosSin1);
  procedure Prod(R: in Flt; P1: in CosSin1; P2: in out CosSin1);
  function "*"(R: Flt; P: CosSin1) return CosSin1;
  procedure AddProd(R: in Flt; P1: in CosSin1; P2: in out CosSin1);
  procedure Mult(Q: in Rational; P: in out CosSin1);
  procedure Mult(S: in Scalar; P: in out CosSin1);
  procedure Prod(S: in Scalar; P1: in CosSin1; P2: in out CosSin1);
  function "*"(S: Scalar; P: CosSin1) return CosSin1;
  procedure AddProd(S: in Scalar; P1: in CosSin1; P2: in out CosSin1);
  procedure Mult(P1: in CosSin1; P2,Tmp: in out CosSin1);
  procedure Mult(P1: in CosSin1; P2: in out CosSin1);
  procedure Prod(P1,P2: in CosSin1; P3: in out CosSin1);
  function "*"(P1,P2: CosSin1) return CosSin1;
  procedure AddProd(P1,P2: in CosSin1; P3: in out CosSin1);
  procedure AddProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1);
  procedure SubProd(P1,P2: in CosSin1; P3: in out CosSin1);
  procedure SubProd(P1,P2: in CosSin1; P3,Tmp: in out CosSin1);
  procedure Div(R: in Flt; P: in out CosSin1);
  procedure Quot(P1: in CosSin1; R: in Flt; P2: in out CosSin1);
  function "/"(P: CosSin1; R: Flt) return CosSin1;

  --- conversion and i/o
  function Get_Precision(P: CosSin1) return Positive;
  procedure Assign(I: in Integer; P: in out CosSin1);  --- sets default radius, parity 0
  procedure Assign(Q: in Rational; P: in out CosSin1); --- sets default radius, parity 0
  procedure Assign(R: in Flt; P: in out CosSin1);      --- sets default radius, parity 0
  procedure Assign(S: in Scalar; P: in out CosSin1);   --- sets default radius, parity 0
  procedure Show1(N: in String; P: in CosSin1; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in CosSin1; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; P: in CosSin1; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out CosSin1; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in CosSin1; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out CosSin1; Decimal: in Boolean := False);

  --- other
  function IsReal(P: CosSin1) return Boolean;
  procedure Real_Part(P: in out CosSin1);
  procedure Imag_Part(P: in out CosSin1);
  procedure Adjoint(P: in out CosSin1);
  procedure Adjoint(P1: in CosSin1; P2: in out CosSin1);
  procedure Simple_Random(P: in out CosSin1);
  procedure Simple_Random(PR: in Radius; PP: in Parity; P: in out CosSin1);
  function MaxNorm(P: CosSin1) return Radius;
  procedure Norm(P: in CosSin1; S: in out Scalar);
  function Norm(P: CosSin1) return Scalar;
  procedure YRange(P: in CosSin1; YInf,YSup: out Flt; Pow2: in Positive := 12);
  function Inf(P: CosSin1; Pow2: Positive := 12) return Flt;
  function Sup(P: CosSin1; Pow2: Positive := 12) return Flt;
  function LInftyNorm(P: CosSin1; Pow2: Positive := 12) return Radius;
  procedure Average(P: in CosSin1; S: in out Scalar);
  function Average(P: CosSin1) return Scalar;
  procedure SetAverage0(P: in out CosSin1);
  procedure SplitOffAverage(P: in out CosSin1; S: in out Scalar);
  procedure Coeff(I: in Natural; P: in CosSin1; S: in out Scalar);
  function Coeff(I: in Natural; P: CosSin1) return Scalar;
  procedure AddCoeff(I: in Natural; S: in Scalar; P: in out CosSin1);
  procedure AddProdToErr(R: in Flt; P1: in CosSin1; P2: in out CosSin1);
  function EffErrTerm(I: Natural; P: CosSin1) return Radius;
  procedure AddErr(I: in Natural; R: in Flt; P: in out CosSin1);
  procedure AddBall(R: in Radius; P: in out CosSin1);
  procedure ValAt0(P: in CosSin1; S: in out Scalar);
  procedure Val(P: in CosSin1; S1: in Scalar; S2: in out Scalar);
  function Val(P: CosSin1; S: Scalar) return Scalar;
  procedure TranslatePi(P: in out CosSin1);
  procedure TranslatePi(P1: in CosSin1; P2: in out CosSin1);
  procedure NegTranslatePi(P1: in CosSin1; P2: in out CosSin1);
  procedure SymmetrizePi(SP: in Parity; P: in out CosSin1; Check: in Boolean := False);
  function Convert(C: Flt_Poly1; First: Natural := 0) return SPoly1;
  procedure Compose(C: in SPoly1; P1: in CosSin1; P2: in out CosSin1);
  function DerConvert(C: Flt_Poly1; First: Natural := 0) return SPoly1;
  function Der(C: SPoly1) return SPoly1;
  procedure Der(P: in out CosSin1; R: in Radius);
  procedure Der(P1: in CosSin1; P2: in out CosSin1; R: in Radius);
  procedure AntiDer(P: in out CosSin1);
  procedure AntiDer(P1: in CosSin1; P2: in out CosSin1);
  function Integral(P: CosSin1; S1,S2: Scalar) return Scalar;
  procedure InvHelm(Omega2,M: in Scalar; P: in out CosSin1);
  procedure InvHelm(Omega2,M: in Scalar; P1: in CosSin1; P2: in out CosSin1);
  function PowersOf(S: Scalar; MaxPower: Positive) return SVector;

  --- numerical
  procedure NumTrunc(Siz: in Natural; P1: in CosSin1; P2: in out CosSin1);
  function NumVal(P: CosSin1; T: Flt) return Flt;
  procedure PutPlot(F: in File_Type; P: in CosSin1; Steps: in Positive := 64; PutX: in Boolean := False);
  function GuessRho(P: CosSin1) return Radius;
  procedure EvenOdd(P: in out CosSin1);
  procedure OddEven(P: in out CosSin1);

  -----------------------------------------------------
  -----------------------------------------------------

  subtype Weights is Errs1;

  protected type Protected_NormFactors is
    procedure Get(R: in Radius; W: out Weights);
  private
    RSaved: Radius := One;
    WSaved: Weights := (others => One);
  end Protected_NormFactors;

  protected type Protected_DerFactors is
    procedure Get(R1,R2: in Radius; Fac: out Weights);
  private
    RSaved1: Radius := One;
    RSaved2: Radius := Zero;
    FacSaved: Weights := (others => Zero);
  end Protected_DerFactors;

  NormFactors: Protected_NormFactors;
  DerFactors: Protected_DerFactors;

  package PAlloc is new Recycling (Item => CosSin1);
  subtype LP_Pointer is PAlloc.LD_Pointer;
  PAC: PAlloc.Controlled_Protected_Data;
  PPool: PAlloc.Protected_Data renames PAC.P;

private

  subtype CAbs is FltVec(FFreq);

  function Is_Zero(P: Parity; C: FPoly1) return Boolean;
  function Is_Zero(E: Errs1) return Boolean;
  procedure LiftE0(P: in out CosSin1);
  procedure CCProd2(P1P,P2P: in Parity; C1,C2: in FPoly1; A1,A2: in CAbs; P3: in out CosSin1);
  procedure CEProd2(P1P: in Parity; A1: in CAbs; E2: in Errs1; E3: in out Errs1; W: in Weights);
  procedure EEProd2A(E1,E2: in Errs1; E3: in out Errs1; W: in Weights);
  procedure EEProd2B(E1,E2: in Errs1; E3: in out Errs1; W: in Weights);
  procedure EEProd2(E1,E2: in Errs1; E3: in out Errs1; W: in Weights) renames EEProd2B;
  procedure Make_NormFactors(R: in Radius; W: out Weights);
  function QuasiDeg(P: Parity; C: FPoly1) return Natural;
  function QuasiDeg1(C: FPoly1) return Natural;
  procedure ValAt0(P: in CosSin1; S: in out Scalar; D: in Natural; Is_Poly: in Boolean);
  procedure Val(P: in CosSin1; S1: in Scalar; S2: in out Scalar; D: in Natural; Is_Poly: in Boolean);
  procedure Compose_Part(C: in SPoly1; P1: in CosSin1; P2,Tmp: in out CosSin1);
  procedure Make_DerFactors(R1,R2: in Radius; Fac: out Weights);
  function InfAbs(S: Scalar) return Radius;

  SZero:      constant Scalar      := Scal(0);
  SHalf:      constant Scalar      := Scal(Half);
  SOne:       constant Scalar      := Scal(1);
  STwo:       constant Scalar      := Scal(2);
  SInfo:      constant Scalar_Info := Info(SZero);
  STrunc:     constant Boolean     := SInfo.IsNumeric;
  Not_STrunc: constant Boolean     := not STrunc;

  PInfo: constant Scalar_Info := (RepType    => SInfo.RepType,
                                  IsStandard => SInfo.IsStandard,
                                  IsNumeric  => SInfo.IsNumeric,
                                  IsComplex  => SInfo.IsComplex,
                                  NReps      => SInfo.NReps*(FDeg+1),
                                  RMode      => SInfo.RMode);

  Default_Radius: Radius := Two;
  Default_Parity: Parity := 0;

end CosSins1;
