with Flts, Globals, Reps, Params, Params.Names, Flok_Intervals, Strings;
use Flts, Globals, Params, Flok_Intervals, Strings;
with Flt_Vectors, Matrices, Vectors, ScalVectors, ScalVectors.Phi;
with Reps.Balls;
use Reps.Balls;

pragma Elaborate_All (Flts,Globals,Reps,Params.Names,Flok_Intervals);
pragma Elaborate_All (Flt_Vectors,Matrices,Vectors,ScalVectors.Phi);
pragma Elaborate_All (Reps.Balls);

procedure Eigen(Sol: in Positive) is

  package PN is new Params.Names(Sol => Sol, Trunc => Scalar_IsNumeric);
  use PN;

  subtype Scalar is Ball;

  NumDigits: constant Positive := Reps.Set_Precision(1024);

  package SM is new Matrices (Scalar => Scalar);
  package SV is new Vectors (Component => Scalar);
  package SSV is new ScalVectors (Scalar => Scalar, SM => SM, SV => SV);
  package SSV_Phi is new SSV.Phi;

  use SM,SV,SSV,SSV_Phi;

  DimX:  constant Positive := 2*JAst+1-Sigma;   ------- size of the truncated chain
  Dim2X: constant Positive := 2*DimX;
  Eps:     constant Radius := 1.0E-28; --- for bisection in EigenCircle

  Wr,Wi: Vector(1..DimX);
  Vr,Vi: Matrix(1..Dim2X,1..DimX);
  A: Matrix(1..Dim2X,1..Dim2X);

  function Sep_Eta(Low,High: F_Intervals) return FltVec is
    S: constant FltVec(1..3) := (Half**4,Half,One-Half**4); --- by inspection
  begin
    if Prefix="EVEN" then return S; end if;
    return Separating_Eta(Low,High);
  end Sep_Eta;

  procedure Eigenvectors_And_More(PM: in Integer) is
    use Flt_FS,Flt_Vectors;
    Name: constant String := Phi_Name(PM);
    NPD,Mu: Radius := Zero;
  begin
    Read(Input_Dir & Name,A);
    EigenSymplectic(A,Wr,Wi,Vr,Vi,Eps,-1);
    Show2("",Wr,Wi);
    PosIndexG(Wr,Wi,Vr,Vi);
    Plot_Eigen_Pos(Output_Dir & Name,Wr,Wi);
    Plot_Eigen_Neg(Output_Dir & Name,Wr,Wi);
    Plot_Eigen_Zero(Output_Dir & Name,Wr);
    if RealEV=0 then
      Col_Normalize1(Vr,Vi);
      if PM=1 then
        declare
          Lo: constant F_Intervals := Low_Eta(Wr,Wi,Vr,Vi);
          Hi: constant F_Intervals := Read(Output_Dir & Hi_Eta_Name);
          Sep: constant FltVec := Sep_Eta(Lo,Hi);
        begin
          Write(Output_Dir & Prefix & "-lo-eta.dec" & Post,Lo,True);
          Write(Output_Dir & Prefix & "-sep-eta" & Post,Sep,False);
          Write(Output_Dir & Prefix & "-sep-eta.dec" & Post,Sep,True);
        end;
      end if;
    else
      Read(Input_Dir & NPD_Name,NPD);
      Mu := FindMu(Wr);
      if Check_Unstable(A,Mu,NPD) then
        Show0("unstable ok");
      else
        Show0("ko");
      end if;
    end if;
  end Eigenvectors_And_More;

begin

  if RealEV=0 then
    Eigenvectors_And_More(-1);
    Eigenvectors_And_More(1);
  else
    Eigenvectors_And_More(0);
  end if;

end Eigen;
