with Ints, Recycling, Flt_Vectors;
use Ints;

pragma Elaborate_All (Ints,Recycling,Flt_Vectors);

generic

  Dim: in Positive;

package MultiTaylors1.Phi is

  Dim2: constant Positive := 2*Dim;
  subtype IndexX is Integer range 1 .. Dim;
  subtype Index2X is Integer range 1 .. Dim2;

  subtype YType is Scalar_Vector(IndexX);
  subtype SMatrix is Scalar_Matrix(Index2X,Index2X);
  subtype SVector is Scalar_Vector(Index2X);
  type SVecVec is array(Index2X) of SVector; -- for Phi00

  type TMatrix is array(Power) of SMatrix;
  type TVector is array(Power) of SVector;
  type TVecVec is array(Index2X) of TVector; -- for Phi

  procedure Show_Param;
  procedure LinFlow(X: in TMatrix; V: in out TVecVec; DT: in Radius);
  procedure Run(DT: in Radius; M1,M2: in Scalar_Matrix; C: in Scalar_Vector; Phi00: in out SMatrix);

  Flow_Parallel: Boolean := True;

private

  subtype Flt_Vector is Flt_Vectors.Vector;

  PDeg1:  constant Positive := PDeg+1;
  EDeg:   constant Positive := Choose(STrunc,PDeg1,PDeg);
  SNegOne: constant Scalar  := Scal(-1);

  subtype TMatrix1 is Taylor1_Matrix(Index2X,Index2X);
  subtype TVector1 is Taylor1_Vector(Index2X);

  function HSymmetry(Xh: TMatrix1) return Radius;
  function Symplecticity(U: Scalar_Matrix) return Radius;
  function MaxNorm(V: SVector) return Radius;
  procedure AddBall(R: in Radius; V: in out SVector);
  procedure Quot(V: in SVector; R: in Flt; W: in out SVector);
  function OpNorm(A: SMatrix) return Radius;
  procedure LinFlow(X: in TMatrix; V: in out TVector; DT: in Radius; Tmp1,Tmp2: in out SVector);

  procedure SetZero(V: in out TVector);
  procedure T2T(A: in out TMatrix1; B: in out TMatrix);
  procedure Val(P: in TVector; X: in Scalar; Y: in out SVector);
  procedure MakeXh(M1,M2: in Scalar_Matrix; C: in Scalar_Vector; Time,DT: in Radius; Xh: in out TMatrix1);
  function MakePhi0(DT: Radius) return TMatrix1;
  function MakeInvPhi0(DT: Radius) return TMatrix1;
  procedure ValPhi0(T: in Scalar; V: in out SMatrix);

  package MAlloc is new Recycling (Item => SVector);
  subtype LM_Pointer is MAlloc.LD_Pointer;
  MAC: MAlloc.Controlled_Protected_Data;
  MPool: MAlloc.Protected_Data renames MAC.P;

end MultiTaylors1.Phi;
