
package body MultiTaylors1 is

  use SV_More,STM;

  procedure SetRho(R: in Radius; P: in out Taylor1_Vector; Check: in Boolean := True) is
  begin
    for N in P'Range loop SetRho(R,P(N),Check); end loop;
  end SetRho;

  function Rho(P: in Taylor1_Vector) return Radius is
    R: Radius := Radius'Last;
  begin
    for N in P'Range loop R := RMin(Rho(P(N)),R); end loop;
    return R;
  end Rho;

  function GuessRho(P: in Taylor1_Vector) return Radius is
    R: Radius := Radius'Last;
  begin
    for N in P'Range loop R := RMin(GuessRho(P(N)),R); end loop;
    return R;
  end GuessRho;

  procedure Components(I: in Natural; P: in Taylor1_Vector; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop Component(I,P(N),W(N)); end loop;
  end Components;

  procedure ZeroComponents(I: in Natural; P: in out Taylor1_Vector) is
  begin
    for N in P'Range loop SetZero(P(N).C(I)); end loop;
  end ZeroComponents;

  procedure AddCoeffs(I: in Natural; W: in Scalar_Vector; P: in out Taylor1_Vector) is
  begin
    for N in P'Range loop AddCoeff(I,W(N),P(N)); end loop;
  end AddCoeffs;

  procedure Val(P: in Taylor1_Vector; S: in Scalar; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop Val(P(N),S,W(N)); end loop;
  end Val;

  procedure ValDer(P: in Taylor1_Vector; S: in Scalar; W: in out Scalar_Vector) is
  begin
    for N in W'Range loop ValDer(P(N),S,W(N)); end loop;
  end ValDer;

  procedure Assign(W: in Scalar_Vector; P: in out Taylor1_Vector) is
  begin
    for N in W'Range loop Assign(W(N),P(N)); end loop;
  end Assign;

  -----------------------

  procedure SetRho(R: in Radius; P: in out Taylor1_Matrix; Check: in Boolean := True) is
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        SetRho(R,P(M,N),Check);
      end loop;
    end loop;
  end SetRho;

  function Rho(P: in Taylor1_Matrix) return Radius is
    R: Radius := Radius'Last;
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        R := RMin(Rho(P(M,N)),R);
      end loop;
    end loop;
    return R;
  end Rho;

  function GuessRho(P: in Taylor1_Matrix) return Radius is
    R: Radius := Radius'Last;
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        R := RMin(GuessRho(P(M,N)),R);
      end loop;
    end loop;
    return R;
  end GuessRho;

  procedure Components(I: in Natural; P: in Taylor1_Matrix; W: in out Scalar_Matrix) is
  begin
    for M in W'Range(1) loop
      for N in W'Range(2) loop
        Component(I,P(M,N),W(M,N));
      end loop;
    end loop;
  end Components;

  procedure ZeroComponents(I: in Natural; P: in out Taylor1_Matrix) is
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        SetZero(P(M,N).C(I));
      end loop;
    end loop;
  end ZeroComponents;

  procedure AddCoeffs(I: in Natural; W: in Scalar_Matrix; P: in out Taylor1_Matrix) is
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        AddCoeff(I,W(M,N),P(M,N));
      end loop;
    end loop;
  end AddCoeffs;

  procedure Assign(W: in Scalar_Matrix; P: in out Taylor1_Matrix) is
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        Assign(W(M,N),P(M,N));
      end loop;
    end loop;
  end Assign;

  procedure Val(P: in Taylor1_Matrix; S: in Scalar; W: in out Scalar_Matrix) is
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        Val(P(M,N),S,W(M,N));
      end loop;
    end loop;
  end Val;

  procedure ValDer(P: in Taylor1_Matrix; S: in Scalar; W: in out Scalar_Matrix) is
  begin
    for M in P'Range(1) loop
      for N in P'Range(2) loop
        ValDer(P(M,N),S,W(M,N));
      end loop;
    end loop;
  end ValDer;

end MultiTaylors1;
