with Ints, ACM_RPoly;
use Ints;

pragma Elaborate_All (Ints,ACM_RPoly);

package body Num_PolyRoots is

  procedure NumRealRoot(Deg: in Positive; WMin,WMax: in Flt; Eps: in Radius) is
    P: Poly(0 .. Deg);
    W: Numeric;
  begin
    Reps.Proper_Rounding;
    for I in 0 .. Deg loop Get_Numeric(I+1,P(I)); end loop;
    NumRealRoot(P,WMin,WMax,W,Eps);
    Put_Numeric(1,W); -- ok if omitted when Not_Found
  end NumRealRoot;

  procedure NumRealRoots(Deg: in Positive; WMin,WMax: in Flt; Eps: in Radius) is
    P: Poly(0 .. Deg);
    W: NVector(1 .. Deg);
  begin
    Reps.Proper_Rounding;
    for I in 0 .. Deg loop Get_Numeric(I+1,P(I)); end loop;
    NumRealRoots(P,WMin,WMax,W,Eps);
    for I in 1 .. Deg loop Put_Numeric(I,W(I)); end loop;
  end NumRealRoots;

  procedure NumSymQFactor(Deg: in Positive; UMin,UMax: in Flt; Eps: in Radius) is
    P: Poly(0 .. Deg);
    U: Numeric;
  begin
    Reps.Proper_Rounding;
    for I in 0 .. Deg loop Get_Numeric(I+1,P(I)); end loop;
    NumSymQFactor(P,UMin,UMax,U,Eps);
    Put_Numeric(1,U); -- ok if omitted when Not_Found
  end NumSymQFactor;

  procedure NumRoots(P: in Poly; Wr,Wi: in out Vector; D: out Natural) is
    Deg: constant Natural := EffDeg(P);
    package RP is new ACM_RPoly (MaxDeg => Deg, Scalar => Numeric, Scalars => NVector);
    use RP;
    Deg1: constant Positive := Deg+1;
    Eps: constant Radius := 1.0E-8;   --- for sorting
    Fail: Boolean;
    OP: Coeffs;
  begin
    D := Deg;
    for I in 0 .. Deg loop Copy(P(I),OP(Deg1-I)); end loop;
    RPoly(OP,D,Wr,Wi,Fail);
    if Fail then
      Show1("P has degree",Deg);
      Show1("the number of roots found is only",D);
      for I in D+1 .. Wr'Last loop
        SetZero(Wr(I));
        SetZero(Wi(I));
      end loop;
    end if;
    Sort(Eps,Wr,Wi);
    FindRoots(P,Wr,Wi,D);
  end NumRoots;

  procedure NumRoots(D: in out Natural) is
    Deg: constant Positive := D;
    P: Poly(0 .. Deg);
    Wr,Wi: NVector(1 .. Deg);
  begin
    Reps.Proper_Rounding;
    for I in 0 .. Deg loop Get_Numeric(I+1,P(I)); end loop;
    NumRoots(P,Wr,Wi,D);
    for I in 1 .. Deg loop Put_Numeric(I,Wr(I)); end loop;
    for I in 1 .. Deg loop Put_Numeric(Deg+I,Wi(I)); end loop;
  end NumRoots;

end Num_PolyRoots;
