with Flts, Ints, Flt_Vectors, Ada.Text_IO, Rationals;
use Flts, Ints, Flt_Vectors, Ada.Text_IO, Rationals;

pragma Elaborate_All (Flts,Ints,Flt_Vectors,Ada.Text_IO,Rationals);

package Params is

  DegPhiPsiPrime: constant Integer := 3;

  subtype Potential is Flt_Vectors.Vector(0..DegPhiPsiPrime);
  type Symmetry is (Symm, ASymm, None);

  type Param is
    record
      SizeChain: Positive;
      SizeCosSin: Positive; -- coefficients for the Fourier expansion
      JAst: Positive;       -- J*
      TDeg: Positive;       -- degree of Taylor polynomials in Phi2PI
      ExDT: Natural;        -- DT in Phi2Pi is 2^(-ExDT); if ExDT=0 then do not compute Phi2Pi
      Sigma: Parity;        -- 0: site-centered  1: bond-centered
      TZ:     Symmetry;     -- reflection with time-translation zero
      TPi:    Symmetry;     -- reflection with time-translation Pi
      Par: Parity;          -- 0 means solutions even in time, 1 odd in time
      Rad: Radius;          -- CosSin1
      Rho: Radius;          -- Chain
      Om:  Rational;        -- omega
      Psi1: Potential;      -- first derivative of \Psi
      Phi1: Potential;      -- first derivative of \Phi
      RealEV: Natural;      -- number of real eigenvalues in (-1,1)
      HatMaxSize: Positive; -- for separation
      Mat_PackSize: Positive;
      Contr_PackSize: Positive;
      Prefix: String(1..4);
    end record;

  SISY: constant Param := (SizeChain => 7,   --- 1  SIte centered SYmmetric
                           SizeCosSin => 80,
                           JAst => 3,
                           TDeg => 50,
                           ExDT => 6,
                           Sigma => 0,
                           TZ => Symm,
                           TPi => None,
                           Par => 1,
                           Rad => Two,
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,Half**20,Zero,Two),
                           Psi1 => (Zero,Half,Zero,Zero),
                           RealEV => 1,
                           HatMaxSize => 1,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "SISY");

  SIAS: constant Param := (SizeChain => 15,   ---- 2 SIte centered ASymmetric
                           SizeCosSin => 120,
                           JAst => 5,
                           Sigma => 0,
                           TDeg => 50,
                           ExDT => 6,
                           TZ => ASymm,
                           TPi => None,
                           Par => 1,
                           Rad => Two,
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,Half**20,Zero,Two),
                           Psi1 => (Zero,Half,Zero,Zero),
                           RealEV => 0,
                           Mat_PackSize => 15,
                           HatMaxSize => 120,
                           Contr_PackSize => 30,
                           Prefix => "SIAS");

  BOSY: constant Param := (SizeChain => 14, -- 3  BOnd centered SYmmetric
                           SizeCosSin => 80,
                           TDeg => 50,
                           ExDT => 6,
                           JAst => 5,
                           Sigma => 1,
                           TZ => Symm,
                           TPi => None,
                           Par => 1,
                           Rad => Two,
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,Half**3,Zero,One),
                           Psi1 => (Zero,Half,Zero,Zero),
                           RealEV => 0,
                           HatMaxSize => 60,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "BOSY");   -- link centered, symmetric, TPi false

  BOAS: constant Param := (SizeChain => 12,  -- 4  BOnd centered ASymmetric
                           SizeCosSin => 80,
                           JAst => 4,
                           TDeg => 50,
                           ExDT => 6,
                           Sigma => 1,
                           TZ => ASymm,
                           TPi => None,
                           Par => 1,
                           Rad => Two,
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,Half**3,Zero,One),
                           Psi1 => (Zero,Half,Zero,Zero),
                           RealEV => 0,
                           HatMaxSize => 60,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "BOAS");

  ARPU: constant Param := (SizeChain => 18, ---- 5 Always Repulsive Phi Unstable
                           SizeCosSin => 80,
                           JAst => 6,
                           TDeg => 1,
                           ExDT => 0,
                           Sigma => 1,
                           TZ => Asymm,
                           TPi => None,
                           Par => 1,
                           Rad => One+Half**2,
                           Rho => One+Half**2,
                           Om => 32/5,
                           Phi1 => (Zero,-One,Zero,-Flt(27)/Flt(16)),
                           Psi1 => (Zero,One,Zero,One),
                           RealEV => 1,
                           Mat_PackSize => 15,
                           HatMaxSize => 1,
                           Contr_PackSize => 30,
                           Prefix => "ARPU");

  NEUA: constant Param := (SizeChain => 10,   -- 6  NEgative phi''(0) Unstable Asymmetric
                           SizeCosSin => 25,
                           JAst => 4,
                           TDeg => 1,
                           ExDT => 0,
                           Sigma => 1,
                           TZ => ASymm,
                           TPi => None,
                           Par => 1,
                           Rad => One+Half**2,
                           Rho => One+Half**2,
                           Om => 256/5,
                           Phi1 => (Zero,-One,Zero,One),
                           Psi1 => (Zero,Zero,Zero,Zero),
                           RealEV => 0,
                           Mat_PackSize => 15,
                           HatMaxSize => 1,
                           Contr_PackSize => 30,
                           Prefix => "NEUA");

  NEUS: constant Param := (SizeChain => 10,   --  7 NEgative phi''(0) Unstable Symmetric
                           SizeCosSin => 25,
                           JAst => 4,
                           TDeg => 1,
                           ExDT => 0,
                           Sigma => 1,
                           TZ => Symm,
                           TPi => None,
                           Par => 1,
                           Rad => One+Half**2,
                           Rho => One+Half**2,
                           Om => 256/5,
                           Phi1 => (Zero,-One,Zero,One),
                           Psi1 => (Zero,Zero,Zero,Zero),
                           RealEV => 0,
                           HatMaxSize => 1,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "NEUS");

  NESA: constant Param := (SizeChain => 15,  -- 8 NEgative phi''(0) Stable Asymmetric
                           SizeCosSin => 100,
                           JAst => 6,
                           TDeg => 50,
                           ExDT => 7,
                           Sigma => 1,
                           TZ => ASymm,
                           TPi => None,
                           Par => 1,
                           Rad => Flt(17)/Flt(16),
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,-(Half**5),Zero,One),
                           Psi1 => (Zero,One,Zero,Half),
                           RealEV => 0,
                           HatMaxSize => 100,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "NESA");

  NESS: constant Param := (SizeChain => 15,   -- 9 NEgative phi''(0) Stable (numerically) Symmetric
                           SizeCosSin => 100,
                           JAst => 4,
                           TDeg => 50,
                           ExDT => 0,
                           Sigma => 1,
                           TZ => Symm,
                           TPi => None,
                           Par => 1,
                           Rad => Flt(17)/Flt(16),
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,-(Half**5),Zero,One),
                           Psi1 => (Zero,One,Zero,Half),
                           RealEV => 0,
                           HatMaxSize => 60,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "NESS");

  LOCH: constant Param := (SizeChain => 15,  -- 10 to make LOCH NESS multibreather solutions, numerically unstable
                           SizeCosSin => 80,
                           JAst => 4,
                           TDeg => 50,
                           ExDT => 0,
                           Sigma => 0,
                           TZ => Symm,
                           TPi => None,
                           Par => 1,
                           Rad => Flt(17)/Flt(16),
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,-(Half**5),Zero,One),
                           Psi1 => (Zero,One,Zero,Half),
                           RealEV => 1,
                           HatMaxSize => 100,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "LOCH");

  ALRU: constant Param := (SizeChain => 5,  -- 11  As Loch but Rigorously Unstable
                           SizeCosSin => 40,
                           JAst => 4,
                           TDeg => 50,
                           ExDT => 6,
                           Sigma => 0,
                           TZ => Symm,
                           TPi => None,
                           Par => 1,
                           Rad => Flt(17)/Flt(16),
                           Rho => Two,
                           Om => 32/5,
                           Phi1 => (Zero,-(Half**20),Zero,One),
                           Psi1 => (Zero,One,Zero,Half),
                           RealEV => 1,
                           HatMaxSize => 60,
                           Mat_PackSize => 15,
                           Contr_PackSize => 30,
                           Prefix => "ALRU");

  MIXD: constant Param := (SizeChain => 20,  ---- 12 different frequencies MIXeD
                           SizeCosSin => 100,
                           JAst => 14,
                           TDeg => 1,
                           ExDT => 0,
                           Sigma => 1,
                           TZ => None,
                           TPi => None,
                           Par => 1,
                           Rad => One+Half**3,
                           Rho => One+Half**3,
                           Om => 32/5,
                           Phi1 => (Zero,One,Zero,One),
                           Psi1 => (Zero,-Half,Zero,Zero),
                           RealEV => 2,
                           HatMaxSize => 1,
                           Mat_PackSize => 30,
                           Contr_PackSize => 50,
                           Prefix => "MIXD");

  EVEN: constant Param := (SizeChain => 14,  ----  13  EVEN in time
                           SizeCosSin => 140,
                           JAst => 4,
                           TDeg => 50,
                           ExDT => 7,
                           Sigma => 1,
                           TZ => ASymm,
                           TPi => None,
                           Par => 0,
                           Rad => One+Half**3,
                           Rho => One+Half**3,
                           Om => 64/9,
                           Phi1 => (Zero,Half**2,Two,Flt(4)),
                           Psi1 => (Zero,Flt(8),Zero,Zero),
                           RealEV => 0,
                           HatMaxSize => 60,
                           Mat_PackSize => 25,
                           Contr_PackSize => 70,
                           Prefix => "EVEN");

  type Param_List is array(Positive range <>) of Param;

  procedure Show(N: in String; P: in Param);
  procedure Show1(N: in String; T: in Symmetry);
  procedure Put(F: in File_Type; T: in Symmetry);
  procedure Get(F: in File_Type; T: out Symmetry);
  function Max_Size_CosSin(P: Param_List) return Positive;
  function Max_Size_Chain(P: Param_List) return Positive;

  PList: constant Param_List(1..13) := (SISY,SIAS,BOSY,BOAS,ARPU,NEUA,NEUS,
                                        NESA,NESS,LOCH,ALRU,MIXD,EVEN);

end Params;
